"""
The Plugin API.

Plugins allow third-party code (e.g. your own Python package) to add functionality
to Betty.
"""

from __future__ import annotations

from functools import update_wrapper
from importlib import metadata
from typing import TYPE_CHECKING, Any, Self, final, override

from betty.definition.cls import ClsDefinition
from betty.definition.human_facing import CountableHumanFacingDefinition
from betty.importlib import fully_qualified_name
from betty.locale.localizable.gettext import _
from betty.machine_name import MachineName, ResolvableMachineName

if TYPE_CHECKING:
    import builtins
    from collections.abc import Collection, Iterable, Iterator, Mapping

    from betty.locale.localizable import (
        CountableLocalizable,
        Localizable,
        ResolvableLocalizable,
    )
    from betty.plugin.discovery import ResolvableDiscovery


class PluginDefinition[BaseClsT](ClsDefinition[BaseClsT]):
    """
    A plugin definition.
    """

    def __init__(self, plugin_id: ResolvableMachineName, /):
        super().__init__()
        self._id = MachineName.resolve(plugin_id)

    @classmethod
    def type(cls) -> PluginTypeDefinition[BaseClsT, Self]:
        """
        The plugin type definition.
        """
        raise NotImplementedError(
            f"{fully_qualified_name(cls)} was not decorated with a {fully_qualified_name(PluginDefinition)} subclass."
        )

    @property
    def id(self) -> MachineName:
        """
        The plugin ID.

        IDs are unique per plugin type:

        - A plugin repository **MUST** at most have a single plugin for any ID.
        - Different plugin repositories **MAY** each have a plugin with the same ID.
        """
        return self._id

    @override
    def _set_cls(self, cls: builtins.type[BaseClsT]) -> None:
        super()._set_cls(cls)
        cls.plugin = staticmethod(update_wrapper(lambda: self, cls.plugin))  # ty:ignore[unresolved-attribute]

    @property
    def reference_label(self) -> Localizable:
        """
        The label to reference this plugin with.
        """
        return _('"{plugin_id}"').format(plugin_id=self.id)

    @property
    def reference_label_with_type(self) -> Localizable:
        """
        The label to reference this plugin with, including the plugin type.
        """
        return _('{plugin_type} "{plugin_id}"').format(
            plugin_type=self.type().label,
            plugin_id=self.id,
        )


@final
class PluginTypeDefinition[BaseClsT, PluginDefinitionT: PluginDefinition](
    CountableHumanFacingDefinition, ClsDefinition[PluginDefinitionT]
):
    """
    A plugin type definition.
    """

    def __init__(
        self,
        plugin_type_id: ResolvableMachineName,
        *,
        label: ResolvableLocalizable,
        label_plural: ResolvableLocalizable,
        label_countable: CountableLocalizable,
        description: ResolvableLocalizable | None = None,
        discovery: Iterable[ResolvableDiscovery[PluginDefinitionT]] | None = None,
    ):
        super().__init__(
            label=label,
            label_plural=label_plural,
            label_countable=label_countable,
            description=description,
        )

        self._id = MachineName.resolve(plugin_type_id)
        self._discovery = () if discovery is None else tuple(discovery)

    @property
    def id(self) -> MachineName:
        """
        The plugin type ID.
        """
        return self._id

    @override
    def _set_cls(self, cls: type[PluginDefinitionT]) -> None:
        super()._set_cls(cls)
        cls.type = staticmethod(update_wrapper(lambda: self, cls.type))  # ty:ignore[invalid-assignment]

    @property
    def discovery(self) -> Collection[ResolvableDiscovery[PluginDefinitionT]]:
        """
        The plugin discoveries for this type.
        """
        return self._discovery


class Plugin[PluginDefinitionT: PluginDefinition]:
    """
    A plugin class.
    """

    @classmethod
    def plugin(cls) -> PluginDefinitionT:
        """
        The plugin definition.
        """
        raise NotImplementedError(
            f"{fully_qualified_name(cls)} was not decorated with a {fully_qualified_name(PluginDefinition)} subclass."
        )


@final
class PluginTypeRepository:
    """
    A repository of available plugin types.
    """

    def __init__(self):
        self._plugin_types: Mapping[str, type[PluginDefinition]] | None = None

    def _get_plugin_types(self) -> Mapping[str, type[PluginDefinition]]:
        if self._plugin_types is None:
            self._plugin_types = {
                plugin.type().id: plugin
                for entry_point in metadata.entry_points(group="betty.plugin")
                if (plugin := entry_point.load())
            }
        return self._plugin_types

    def __contains__(self, value: Any) -> bool:
        return value in self._get_plugin_types()

    def __getitem__(
        self, plugin_type_id: ResolvableMachineName, /
    ) -> type[PluginDefinition]:
        return self._get_plugin_types()[plugin_type_id]

    def __iter__(self) -> Iterator[type[PluginDefinition]]:
        return iter(self._get_plugin_types().values())


type ResolvableDefinition[PluginDefinitionT: PluginDefinition = PluginDefinition] = (
    PluginDefinitionT | type[Plugin[PluginDefinitionT]]
)
"""
Use :py:func:`betty.plugin.resolve.resolve_definition` to resolve this to a :py:class:`betty.plugin.PluginDefinition`
"""


type ResolvableId[PluginDefinitionT: PluginDefinition = PluginDefinition] = (
    ResolvableMachineName | ResolvableDefinition[PluginDefinitionT]
)
"""
Use :py:func:`betty.plugin.resolve.resolve_id` to resolve this to a plugin ID.
"""


def resolve_definition[PluginDefinitionT: PluginDefinition](
    definition: ResolvableDefinition[PluginDefinitionT], /
) -> PluginDefinitionT:
    """
    Resolve a plugin definition.
    """
    if isinstance(definition, PluginDefinition):
        return definition  # ty:ignore[invalid-return-type]
    return definition.plugin()


def resolve_id(plugin_id: ResolvableId, /) -> MachineName:
    """
    Resolve a plugin identifier to a plugin ID.
    """
    if isinstance(plugin_id, MachineName):
        return plugin_id
    if isinstance(plugin_id, str):
        return MachineName.resolve(plugin_id)
    return resolve_definition(plugin_id).id



"""
Data types representing files on disk.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, final, override

from betty.ancestry.description import HasDescription
from betty.ancestry.has_citations import HasCitations
from betty.ancestry.has_links import HasLinks
from betty.ancestry.has_notes import HasNotes
from betty.ancestry.media_type import HasMediaType
from betty.copyright_notice import CopyrightNoticeDefinition
from betty.license import LicenseDefinition
from betty.locale.localizable.gettext import _, ngettext
from betty.model import EntityDefinition
from betty.model.association import BidirectionalToManyMultipleTypes, ToManyAssociates
from betty.plugin.schema import PluginIdSchema
from betty.privacy import HasPrivacy, Privacy

if TYPE_CHECKING:
    from collections.abc import MutableSequence
    from pathlib import Path

    from betty.ancestry.citation import Citation
    from betty.ancestry.file_reference import FileReference  # noqa: F401
    from betty.ancestry.link import Link
    from betty.ancestry.note import Note
    from betty.copyright_notice import CopyrightNotice
    from betty.json.linked_data import JsonLdObject
    from betty.license import License
    from betty.locale.localizable import Localizable, ResolvableLocalizable
    from betty.media_type import MediaType
    from betty.portable import PortableMapping
    from betty.project import Project


@final
@EntityDefinition(
    "file",
    label=_("File"),
    label_plural=_("Files"),
    label_countable=ngettext("{count} file", "{count} files"),
)
class File(
    HasDescription,
    HasPrivacy,
    HasLinks,
    HasMediaType,
    HasNotes,
    HasCitations,
):
    """
    .. plugin:: entity:file.
    """

    referees = BidirectionalToManyMultipleTypes["File", "FileReference"](
        "betty.ancestry.file_reference:FileReference",
        "file",
        label=_("Referees"),
        description=_("The entities referencing this file"),
        linked_data_embedded=True,
    )
    """
    Other entities referencing this file.
    """

    copyright_notice: CopyrightNotice | None
    """
    The copyright notice for this file.
    """

    license: License | None
    """
    The license for this file.
    """

    def __init__(
        self,
        path: Path,
        *,
        id: str | None = None,  # noqa: A002
        name: str | None = None,
        media_type: MediaType | None = None,
        description: ResolvableLocalizable | None = None,
        notes: ToManyAssociates[Note] | None = None,
        citations: ToManyAssociates[Citation] | None = None,
        privacy: Privacy | None = None,
        links: MutableSequence[Link] | None = None,
        copyright_notice: CopyrightNotice | None = None,
        license: License | None = None,  # noqa: A002
    ):
        super().__init__(
            id,
            media_type=media_type,
            description=description,
            notes=notes,
            citations=citations,
            privacy=privacy,
            links=links,
        )
        self._path = path
        self._name = name
        self.copyright_notice = copyright_notice
        self.license = license

    @property
    def name(self) -> str:
        """
        The file name.
        """
        return self._name or self.path.name

    @property
    def path(self) -> Path:
        """
        The file's path on disk.
        """
        return self._path

    @override
    @property
    def label(self) -> Localizable:
        return self.description or super().label

    @override
    @classmethod
    async def linked_data_schema(cls, project: Project, /) -> JsonLdObject:
        schema = await super().linked_data_schema(project)
        schema.add_property(
            "copyrightNotice",
            PluginIdSchema(
                CopyrightNoticeDefinition.type(),
                await project.plugins.plugins(CopyrightNoticeDefinition),
            ),
            False,
        )
        schema.add_property(
            "license",
            PluginIdSchema(
                LicenseDefinition.type(),
                await project.plugins.plugins(LicenseDefinition),
            ),
            False,
        )
        return schema

    @override
    async def dump_linked_data(self, project: Project, /) -> PortableMapping:
        portable = await super().dump_linked_data(project)
        if self.copyright_notice:
            portable["copyrightNotice"] = self.copyright_notice.plugin().id
        if self.license:
            portable["license"] = self.license.plugin().id
        return portable



"""
Localizable attributes.
"""

from __future__ import annotations

from typing import final

from betty.locale.localizable import (
    CountableLocalizable,
    Localizable,
    ResolvableCountableLocalizable,
    ResolvableLocalizable,
    resolve_countable_localizable,
    resolve_localizable,
)
from betty.locale.localizable.data import (
    CountableLocalizableDefinition,
    LocalizableDefinition,
)
from betty.property import Property


@final
class LocalizableProperty(Property[Localizable, ResolvableLocalizable]):
    """
    A property containing a :py:class:`betty.locale.localizable.Localizable`.
    """

    def __init__(
        self,
        *,
        label: ResolvableLocalizable,
        description: ResolvableLocalizable | None = None,
    ):
        super().__init__(
            LocalizableDefinition(),
            label=label,
            description=description,
            resolver=resolve_localizable,
        )


@final
class CountableLocalizableProperty(
    Property[CountableLocalizable, ResolvableCountableLocalizable]
):
    """
    A property containing a :py:class:`betty.locale.localizable.CountableLocalizable`.
    """

    def __init__(
        self,
        *,
        label: ResolvableLocalizable,
        description: ResolvableLocalizable | None = None,
    ):
        super().__init__(
            CountableLocalizableDefinition(),
            label=label,
            description=description,
            resolver=resolve_countable_localizable,
        )



"""
Provide Betty's ancestry event types.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, final, override

from betty.classtools import Singleton
from betty.event_type import (
    EventType,
    EventTypeDefinition,
    ShouldExistEventType,
)
from betty.locale.localizable.gettext import _, ngettext

if TYPE_CHECKING:
    from betty.ancestry.person import Person
    from betty.project import Project


@final
@EventTypeDefinition(
    "unknown",
    label=_("Unknown"),
    label_plural=_("Unknowns"),
    label_countable=ngettext("{count} unknown", "{count} unknowns"),
)
class Unknown(EventType, Singleton):
    """
    .. plugin:: event-type:unknown.
    """


@final
@EventTypeDefinition(
    "birth",
    label=_("Birth"),
    label_plural=_("Births"),
    label_countable=ngettext("{count} birth", "{count} births"),
)
class Birth(ShouldExistEventType):
    """
    .. plugin:: event-type:birth.
    """

    @override
    @classmethod
    async def should_exist(cls, project: Project, person: Person) -> bool:
        return True


@final
@EventTypeDefinition(
    "death",
    label=_("Death"),
    label_plural=_("Deaths"),
    label_countable=ngettext("{count} death", "{count} deaths"),
    comes_after={Birth},
)
class Death(ShouldExistEventType):
    """
    .. plugin:: event-type:death.
    """

    @override
    @classmethod
    async def should_exist(cls, project: Project, person: Person) -> bool:
        return project.privatizer.has_expired(person, 1)


@final
@EventTypeDefinition(
    "baptism",
    label=_("Baptism"),
    label_plural=_("Baptisms"),
    label_countable=ngettext("{count} baptism", "{count} baptisms"),
    comes_before={Death},
    comes_after={Birth},
    indicates=Birth,
)
class Baptism(EventType):
    """
    .. plugin:: event-type:baptism.
    """


@final
@EventTypeDefinition(
    "bar-mitzvah",
    label=_("Bar mitzvah"),
    label_plural=_("Bar mitzvahs"),
    label_countable=ngettext("{count} bar mitzvah", "{count} bar mitzvahs"),
    comes_before={Death},
    comes_after={Birth},
    indicates=Birth,
)
class BarMitzvah(EventType):
    """
    .. plugin:: event-type:bar-mitzvah.
    """


@final
@EventTypeDefinition(
    "bat-mitzvah",
    label=_("Bat mitzvah"),
    label_plural=_("Bat mitzvahs"),
    label_countable=ngettext("{count} bat mitzvah", "{count} bat mitzvahs"),
    comes_before={Death},
    comes_after={Birth},
    indicates=Birth,
)
class BatMitzvah(EventType):
    """
    .. plugin:: event-type:bat-mitzvah.
    """


@final
@EventTypeDefinition(
    "adoption",
    label=_("Adoption"),
    label_plural=_("Adoptions"),
    label_countable=ngettext("{count} adoption", "{count} adoptions"),
    comes_before={Death},
    comes_after={Birth},
)
class Adoption(EventType):
    """
    .. plugin:: event-type:adoption.
    """


@final
@EventTypeDefinition(
    "funeral",
    label=_("Funeral"),
    label_plural=_("Funerals"),
    label_countable=ngettext("{count} funeral", "{count} funerals"),
    comes_after={Death},
    indicates=Death,
)
class Funeral(EventType):
    """
    .. plugin:: event-type:funeral.
    """


@final
@EventTypeDefinition(
    "cremation",
    label=_("Cremation"),
    label_plural=_("Cremations"),
    label_countable=ngettext("{count} cremation", "{count} cremations"),
    comes_after={Death},
    indicates=Death,
)
class Cremation(EventType):
    """
    .. plugin:: event-type:cremation.
    """


@final
@EventTypeDefinition(
    "burial",
    label=_("Burial"),
    label_plural=_("Burials"),
    label_countable=ngettext("{count} burial", "{count} burials"),
    comes_after={Death},
    indicates=Death,
)
class Burial(EventType):
    """
    .. plugin:: event-type:burial.
    """


@final
@EventTypeDefinition(
    "will",
    label=_("Will"),
    label_plural=_("Wills"),
    label_countable=ngettext("{count} will", "{count} wills"),
    comes_after={Death},
)
class Will(EventType):
    """
    .. plugin:: event-type:will.
    """


@final
@EventTypeDefinition(
    "engagement",
    label=_("Engagement"),
    label_plural=_("Engagements"),
    label_countable=ngettext("{count} engagement", "{count} engagements"),
    comes_after={Birth},
    comes_before={Death},
)
class Engagement(EventType):
    """
    .. plugin:: event-type:engagement.
    """


@final
@EventTypeDefinition(
    "marriage",
    label=_("Marriage"),
    label_plural=_("Marriages"),
    label_countable=ngettext("{count} marriage", "{count} marriages"),
    comes_after={Birth, Engagement},
    comes_before={Death},
)
class Marriage(EventType):
    """
    .. plugin:: event-type:marriage.
    """


@final
@EventTypeDefinition(
    "marriage-announcement",
    label=_("Announcement of marriage"),
    label_plural=_("Announcements of marriage"),
    label_countable=ngettext(
        "{count} announcement of marriage", "{count} announcements of marriage"
    ),
    comes_after={Birth},
    comes_before={Death, Marriage},
)
class MarriageAnnouncement(EventType):
    """
    .. plugin:: event-type:marriage-announcement.
    """


@final
@EventTypeDefinition(
    "divorce",
    label=_("Divorce"),
    label_plural=_("Divorces"),
    label_countable=ngettext("{count} divorce", "{count} divorces"),
    comes_after={Birth, Marriage},
    comes_before={Death},
)
class Divorce(EventType):
    """
    .. plugin:: event-type:divorce.
    """


@final
@EventTypeDefinition(
    "divorce-announcement",
    label=_("Announcement of divorce"),
    label_plural=_("Announcements of divorce"),
    label_countable=ngettext(
        "{count} announcement of divorce", "{count} announcements of divorce"
    ),
    comes_after={Birth, Marriage},
    comes_before={Death, Divorce},
)
class DivorceAnnouncement(EventType):
    """
    .. plugin:: event-type:divorce-announcement.
    """


@final
@EventTypeDefinition(
    "residence",
    label=_("Residence"),
    label_plural=_("Residences"),
    label_countable=ngettext("{count} residence", "{count} residences"),
    comes_after={Birth},
    comes_before={Death},
)
class Residence(EventType):
    """
    .. plugin:: event-type:residence.
    """


@final
@EventTypeDefinition(
    "immigration",
    label=_("Immigration"),
    label_plural=_("Immigrations"),
    label_countable=ngettext("{count} immigration", "{count} immigrations"),
    comes_after={Birth},
    comes_before={Death},
)
class Immigration(EventType):
    """
    .. plugin:: event-type:immigration.
    """


@final
@EventTypeDefinition(
    "emigration",
    label=_("Emigration"),
    label_plural=_("Emigrations"),
    label_countable=ngettext("{count} emigration", "{count} emigrations"),
    comes_after={Birth},
    comes_before={Death},
)
class Emigration(EventType):
    """
    .. plugin:: event-type:emigration.
    """


@final
@EventTypeDefinition(
    "occupation",
    label=_("Occupation"),
    label_plural=_("Occupations"),
    label_countable=ngettext("{count} occupation", "{count} occupations"),
    comes_after={Birth},
    comes_before={Death},
)
class Occupation(EventType):
    """
    .. plugin:: event-type:occupation.
    """


@final
@EventTypeDefinition(
    "retirement",
    label=_("Retirement"),
    label_plural=_("Retirements"),
    label_countable=ngettext("{count} retirement", "{count} retirements"),
    comes_after={Birth},
    comes_before={Death},
)
class Retirement(EventType):
    """
    .. plugin:: event-type:retirement.
    """


@final
@EventTypeDefinition(
    "correspondence",
    label=_("Correspondence"),
    label_plural=_("Correspondences"),
    label_countable=ngettext("{count} correspondence", "{count} correspondences"),
)
class Correspondence(EventType):
    """
    .. plugin:: event-type:correspondence.
    """


@final
@EventTypeDefinition(
    "confirmation",
    label=_("Confirmation"),
    label_plural=_("Confirmations"),
    label_countable=ngettext("{count} confirmation", "{count} confirmations"),
    comes_after={Birth},
    comes_before={Death},
)
class Confirmation(EventType):
    """
    .. plugin:: event-type:confirmation.
    """


@final
@EventTypeDefinition(
    "missing",
    label=_("Missing"),
    label_plural=_("Missings"),
    label_countable=ngettext("{count} missing", "{count} missings"),
    comes_after={Birth},
    comes_before={Death},
)
class Missing(EventType):
    """
    .. plugin:: event-type:missing.
    """


@final
@EventTypeDefinition(
    "conference",
    label=_("Conference"),
    label_plural=_("Conferences"),
    label_countable=ngettext("{count} conference", "{count} conferences"),
    comes_before={Death},
    comes_after={Birth},
)
class Conference(EventType):
    """
    .. plugin:: event-type:conference.
    """



"""
Service levels.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Final

from betty.plugin.manager.service import ServiceLevelPluginManager

if TYPE_CHECKING:
    from betty.plugin.manager import PluginManager
    from betty.service.factory import Factory


class ServiceLevel:
    """
    A service level.
    """

    def __init__(self, *args: Any, plugins: PluginManager | None = None, **kwargs: Any):
        from betty.service.factory import Factory

        super().__init__(*args, **kwargs)
        self._factory = Factory(self)
        self._plugins = ServiceLevelPluginManager(self) if plugins is None else plugins

    @property
    def factory(self) -> Factory:
        """
        The object factory.
        """
        return self._factory

    @property
    def plugins(self) -> PluginManager:
        """
        The plugin manager.
        """
        return self._plugins


UNIVERSE: Final[ServiceLevel] = ServiceLevel()
"""
The universal service level.
"""



"""
Provide plugin repositories for service levels.
"""

from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING, Any, cast, final, override

from betty import plugin
from betty.concurrent import AsynchronizedLock, Ledger
from betty.plugin import PluginDefinition
from betty.plugin.manager import PluginManager
from betty.plugin.repository.static import StaticPluginRepository

if TYPE_CHECKING:
    from collections.abc import MutableMapping

    from betty.machine_name import ResolvableMachineName
    from betty.plugin.repository import PluginRepository
    from betty.service.level import ServiceLevel


@final
class ServiceLevelPluginManager(PluginManager):
    """
    Manage plugins for a service level.
    """

    def __init__(self, services: ServiceLevel, /):
        self._services = services
        self._types = plugin.PluginTypeRepository()
        self._plugin_repositories: MutableMapping[
            type[PluginDefinition], PluginRepository[Any] | None
        ] = defaultdict(None)
        self._ledger = Ledger(AsynchronizedLock.new_threadsafe())

    @override
    @property
    def types(self) -> plugin.PluginTypeRepository:
        return self._types

    @override
    async def plugins[PluginDefinitionT: PluginDefinition](
        self, plugin_type: type[PluginDefinitionT] | ResolvableMachineName, /
    ) -> PluginRepository[PluginDefinitionT]:
        if isinstance(plugin_type, str):
            plugin_type = cast(type[PluginDefinitionT], self.types[plugin_type])
        repository: PluginRepository[PluginDefinitionT] | None
        # If the repository exists already, return it immediately so we avoid acquiring locks.
        repository = self._get(plugin_type)
        if repository:
            return repository
        async with self._ledger.ledger(f"{plugin_type.type().id}"):
            # The repository may have been created since we first checked.
            repository = self._get(plugin_type)
            if repository:
                return repository
            repository = await self._new(plugin_type)
            self._plugin_repositories[plugin_type] = repository
            return repository

    def _get[PluginDefinitionT: PluginDefinition](
        self, plugin_type: type[PluginDefinitionT]
    ) -> PluginRepository[PluginDefinitionT] | None:
        if plugin_type not in self._plugin_repositories:
            return None
        return self._plugin_repositories[plugin_type]

    async def _new[PluginDefinitionT: PluginDefinition](
        self, plugin_type: type[PluginDefinitionT]
    ) -> PluginRepository[PluginDefinitionT]:
        from betty.plugin.discovery import discover

        return StaticPluginRepository(
            plugin_type, *await discover(self._services, *plugin_type.type().discovery)
        )



"""
Discover plugins defined as distribution package entry points.
"""

from __future__ import annotations

from importlib import metadata
from typing import TYPE_CHECKING, cast, final, override

from betty.plugin import PluginDefinition, ResolvableDefinition
from betty.plugin.discovery import PluginDiscovery

if TYPE_CHECKING:
    from collections.abc import Iterable

    from betty.service.level import ServiceLevel


@final
class EntryPointDiscovery[PluginDefinitionT: PluginDefinition](
    PluginDiscovery[PluginDefinitionT]
):
    """
    Discover plugins defined as distribution package `entry points <https://packaging.python.org/en/latest/specifications/entry-points/>`_.

    If you are developing a plugin for an existing plugin type that uses entry points, you'll have
    to add that plugin to your package metadata. For example, for a plugin type

    - whose entry point group is ``your_entry_point_group``
    - with a plugin class ``MyPlugin`` in the module ``my_package.my_module``
    - and a plugin ID ``my-package-plugin``:

    .. code-block:: toml

        [project.entry-points.'betty.your_entry_point_group']
        'my-package-plugin' = 'my_package.my_module:MyPlugin'
    """

    def __init__(
        self,
        entry_point_group: str,
        /,
    ):
        self._entry_point_group = entry_point_group

    @override
    async def discover(
        self, services: ServiceLevel, /
    ) -> Iterable[ResolvableDefinition[PluginDefinitionT]]:
        return [
            cast(ResolvableDefinition[PluginDefinitionT], entry_point.load())
            for entry_point in metadata.entry_points(group=self._entry_point_group)
        ]



"""
Job scheduling.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Awaitable, Callable, Sequence
from typing import TYPE_CHECKING, Self, final

from betty.job import Context

if TYPE_CHECKING:
    from collections.abc import AsyncIterator
    from types import TracebackType

    from betty.job import Job

type ScheduledJobBatch = Callable[[], Awaitable[None]]
"""
A callable to call one or more jobs.

The callable MUST cancel the scheduler if an error is raised.
"""


class Closed(Exception):
    """
    Raised when a scheduler is closed.
    """


class Cancelled(Closed):
    """
    Raised when a scheduler has cancelled.
    """


class Released(Exception):
    """
    Raised when a scheduler has been released.
    """


@final
class CyclicDependencyError(Cancelled):
    """
    Raised when a scheduler has cancelled due to a cyclic dependency.
    """

    def __init__(self, job_ids: Sequence[str], /):
        assert job_ids
        cycle = " -> ".join(f'"{job_id}"' for job_id in job_ids)
        super().__init__(f'Job "{job_ids[0]}" has cyclic dependencies: {cycle}.')


@final
class UnknownJobError(Cancelled):
    """
    Raised when a scheduler has cancelled due to an unknown job.
    """

    def __init__(self, job_id: str, /):
        super().__init__(f'Job "{job_id}" was never added.')


@final
class DuplicateJobError(Cancelled):
    """
    Raised when a scheduler cannot add the same job (ID) more than once.
    """

    def __init__(self, job_id: str, /):
        super().__init__(
            f'Job "{job_id}" was added already, and cannot be added again.'
        )


@final
class Completed(Closed):
    """
    Raised when a scheduler has completed.
    """


class Scheduler(ABC):
    """
    A job scheduler.
    """

    def __init__(self, context: Context | None = None, /):
        self._context = Context() if context is None else context

    @property
    def context(self) -> Context:
        """
        The context for all jobs in this scheduler.
        """
        return self._context

    @abstractmethod
    async def add(self, *jobs: Job) -> None:
        """
        Add a new job.
        """

    @abstractmethod
    async def release(self) -> None:
        """
        Release the scheduler.

        Once called, jobs are released by :py:meth:`betty.job.scheduler.Scheduler.get`, and
        new jobs with dependents can no longer be added.
        """

    @abstractmethod
    async def get(self) -> ScheduledJobBatch:
        """
        Get a batch of jobs to execute.
        """

    async def __anext__(self) -> ScheduledJobBatch:
        try:
            return await self.get()
        except Completed:
            raise StopAsyncIteration from None

    def __aiter__(self) -> AsyncIterator[ScheduledJobBatch]:
        return self

    @abstractmethod
    async def cancel(self, reason: BaseException | None = None, /) -> None:
        """
        Close the scheduler and cancel any pending jobs.
        """

    @abstractmethod
    async def complete(self) -> None:
        """
        Close the scheduler and wait for any pending jobs to complete.
        """

    async def __aenter__(self) -> Self:
        await self.release()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        if exc_val is None:
            await self.complete()
        else:
            await self.cancel(exc_val)



"""
Place type configuration.
"""

from __future__ import annotations

from typing import final, override

from betty.data.aggregate.record.object import ObjectDefinition
from betty.locale import DEFAULT_LOCALE
from betty.locale.localizable.gettext import _
from betty.locale.localizable.static import CountableStaticTranslations
from betty.place_type import PlaceType, PlaceTypeDefinition
from betty.plugin.data import CountableHumanFacingPluginDefinitionConfiguration
from betty.sample import Sample


@final
@ObjectDefinition(
    label=_("Place type configuration"),
    samples=[
        lambda: Sample(
            PlaceTypeDefinitionConfiguration(
                id="moon",
                label="Moon",
                label_plural="Moons",
                label_countable=CountableStaticTranslations(
                    {
                        DEFAULT_LOCALE: {
                            "one": "{count} moon",
                            "other": "{count} moons",
                        }
                    }
                ),
            ),
            label="Default",
        )
    ],
)
class PlaceTypeDefinitionConfiguration(
    CountableHumanFacingPluginDefinitionConfiguration[PlaceTypeDefinition]
):
    """
    Configure a :py:class:`betty.place_type.PlaceTypeDefinition`.

    .. data:: betty.project.data:PlaceTypeDefinitionConfiguration
    """

    @override
    def new_plugin(self) -> PlaceTypeDefinition:
        @PlaceTypeDefinition(
            self.id,
            label=self.label,
            label_plural=self.label_plural,
            label_countable=self.label_countable,
            description=self.description,
        )
        class _ProjectConfigurationPlaceType(PlaceType):
            pass

        return _ProjectConfigurationPlaceType.plugin()



"""
Tools to automatically provide repositories for plugin types.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from betty.plugin import PluginDefinition, PluginTypeRepository
from betty.typing import threadsafe

if TYPE_CHECKING:
    from betty.machine_name import ResolvableMachineName
    from betty.plugin.repository import PluginRepository


@threadsafe
class PluginManager(ABC):
    """
    Find plugin types and plugins.
    """

    @abstractmethod
    async def plugins[PluginDefinitionT: PluginDefinition](
        self, plugin_type: type[PluginDefinitionT] | ResolvableMachineName, /
    ) -> PluginRepository[PluginDefinitionT]:
        """
        Get the available plugins for the given type.
        """

    @property
    @abstractmethod
    def types(self) -> PluginTypeRepository:
        """
        Get the available plugin types.
        """



"""
Dynamic content.
"""

from __future__ import annotations

from asyncio import gather
from collections.abc import Iterable, Mapping, Sequence
from typing import TYPE_CHECKING, Self, final, override

from betty.ancestry.citation import Citation
from betty.ancestry.event import Event
from betty.ancestry.file import File
from betty.ancestry.has_citations import HasCitations
from betty.ancestry.has_file_references import HasFileReferences
from betty.ancestry.has_links import HasLinks
from betty.ancestry.person import Person
from betty.ancestry.place import Place
from betty.ancestry.source import Source
from betty.assertion import (
    assert_enum,
    assert_int,
    assert_mapping,
    assert_or,
    assert_sequence,
)
from betty.content import (
    Content,
    ContentDefinition,
    ContentManufacturer,
    build,
)
from betty.content.contents import (
    Render,
    RenderConfiguration,
    Template,
    TemplateBuild,
)
from betty.data import Data, Sample, Size
from betty.data.aggregate.collection.mapping import MappingDefinition
from betty.data.aggregate.collection.sequence import SequenceDefinition
from betty.data.aggregate.record.object import ObjectDefinition
from betty.data.bool import BoolDefinition
from betty.data.enum import EnumDefinition
from betty.data.int import IntDefinition
from betty.extension._theme import (
    associated_file_references,
    person_timeline_events,
    place_timeline_events,
)
from betty.extension.raspberry_mint import (
    Breakpoint,
    JustifyContent,
    RaspberryMint,
)
from betty.extension.raspberry_mint import ColorStyle as RaspberryMintColorStyle
from betty.functools import unique
from betty.locale.localizable.gettext import _
from betty.locale.localizable.property import LocalizableProperty
from betty.machine_name import MachineName, MachineNameProperty, ResolvableMachineName
from betty.model.reference import EntityReference
from betty.plugin import resolve_id
from betty.plugin.data import PluginManufacturerSequenceDefinition
from betty.plugin.data.property import PluginManufacturerSequenceProperty
from betty.portable import CallbackPorter
from betty.privacy import is_public
from betty.property import Optional, Property
from betty.role import RoleDefinition
from betty.service.factory import DataManufacturable, Manufacturable
from betty.service.requirement.extension import require_extension
from betty.service.requirement.project import require_project
from betty.test_utils.locale.localizable import DUMMY_LOCALIZABLE

if TYPE_CHECKING:
    from betty.ancestry import Ancestry
    from betty.document import Document
    from betty.jinja import Environment
    from betty.locale.localizable import ResolvableLocalizable
    from betty.model import Entity
    from betty.plugin import ResolvableId
    from betty.plugin.factory import ResolvablePluginManufacturerSequence
    from betty.project import Project


@final
@ObjectDefinition(
    label=_("Section configuration"),
    samples=[
        lambda: Sample(
            SectionConfiguration(
                ContentManufacturer("my-first-content"),
                heading=DUMMY_LOCALIZABLE,
            ),
            label="Minimal",
            size=Size.MINIMAL,
        ),
    ],
)
class SectionConfiguration(Data):
    """
    Configuration for :py:class:`betty.extension.raspberry_mint.content.Section`.

    .. data:: betty.extension.raspberry_mint.content:SectionConfiguration
    """

    content = PluginManufacturerSequenceProperty[ContentDefinition, Content](
        ContentManufacturer, label=_("Content")
    )
    """
    The content within this section.
    """

    heading = LocalizableProperty(label=_("Heading"))
    """
    The section heading.
    """

    name = Optional(MachineNameProperty())
    """
    The section's machine name, used to generate permanent links.
    """

    visually_hide_heading = Optional(
        Property(
            BoolDefinition(label=_("Visually hide heading")),
            omit_load=True,
            omit_dump=lambda data: data is False,
        )
    )
    """
    Visually hide the heading.
    
    This keeps the heading for accessibility purposes, but does not display it visually.
    """

    def __init__(
        self,
        content: ResolvablePluginManufacturerSequence[ContentDefinition, Content],
        *,
        heading: ResolvableLocalizable,
        name: ResolvableMachineName | None = None,
        visually_hide_heading: bool | None = None,
    ):
        super().__init__()
        self.content = content
        self.heading = heading
        self.name = name
        self.visually_hide_heading = visually_hide_heading


@ContentDefinition("raspberry-mint-section", label=_("Section"))
class Section(Template, DataManufacturable[SectionConfiguration]):
    """
    .. plugin:: content:raspberry-mint-section.
    """

    def __init__(
        self,
        /,
        content: Iterable[Content],
        *,
        heading: ResolvableLocalizable,
        name: MachineName | None = None,
        visually_hide_heading: bool | None = None,
        jinja: Environment,
    ):
        super().__init__(jinja=jinja)
        self._content = tuple(content)
        self._heading = heading
        self._name = name
        self._visually_hide_heading = bool(visually_hide_heading)

    @override
    @classmethod
    def new_data_cls(cls) -> type[SectionConfiguration]:
        return SectionConfiguration

    @override
    @classmethod
    @require_project
    async def new(cls, project: Project, data: SectionConfiguration, /) -> Self:
        await require_extension(RaspberryMint, project)
        content, jinja = await gather(
            gather(
                *map(
                    project.factory.new,
                    map(ContentManufacturer.resolve, data.content),
                )
            ),
            project.jinja,
        )
        return cls(
            content=content,
            heading=data.heading,
            jinja=jinja,
            name=data.name,
            visually_hide_heading=data.visually_hide_heading,
        )

    @override
    async def build_template(self, document: Document) -> TemplateBuild:
        content = await build(document, self._content)
        if content is None:
            return None
        return "component/raspberry-mint/section.html.j2", {
            "section_content": content,
            "section_heading": self._heading,
            "section_name": self._name,
            "section_visually_hide_heading": self._visually_hide_heading,
        }


@ContentDefinition("raspberry-mint-entity-card", label=_("Entity card"))
class EntityCard(Template, DataManufacturable[EntityReference]):
    """
    A card featuring an entity.

    .. plugin:: content:raspberry-mint-entity-card
    """

    def __init__(
        self, *, ancestry: Ancestry, entity: EntityReference, jinja: Environment
    ):
        super().__init__(jinja=jinja)
        self._entity = entity
        self._ancestry = ancestry

    @override
    @classmethod
    def new_data_cls(cls) -> type[EntityReference]:
        return EntityReference

    @override
    @classmethod
    @require_project
    async def new(cls, project: Project, data: EntityReference, /) -> Self:
        await require_extension(RaspberryMint, project)
        return cls(
            ancestry=project.ancestry,
            entity=data,
            jinja=await project.jinja,
        )

    @override
    async def build_template(self, document: Document) -> TemplateBuild:
        entity = self._ancestry[self._entity.type][self._entity.id]
        return [
            "entity/card--" + entity.plugin().id + ".html.j2",
            "entity/card.html.j2",
        ], {"entity": entity}


@ContentDefinition("raspberry-mint-families", label=_("Families"))
class Families(Template, Manufacturable):
    """
    A person's families.

    .. plugin:: content:raspberry-mint-families
    """

    @override
    @classmethod
    @require_project
    async def new(cls, project: Project, /) -> Self:
        await require_extension(RaspberryMint, project)
        return cls(jinja=await project.jinja)

    @override
    async def build_template(self, document: Document) -> TemplateBuild:
        if isinstance(document.resource, Person):
            return "component/raspberry-mint/families.html.j2", {
                "person": document.resource,
            }
        return None


@ContentDefinition(
    "raspberry-mint-media",
    label=_("Media"),
    description=_("A single file in a media display"),
)
class Media(Template, Manufacturable):
    """
    A single file in a media display.

    .. plugin:: content:raspberry-mint-media
    """

    @override
    @classmethod
    @require_project
    async def new(cls, project: Project, /) -> Self:
        await require_extension(RaspberryMint, project)
        return cls(jinja=await project.jinja)

    @override
    async def build_template(self, document: Document) -> TemplateBuild:
        if isinstance(document.resource, File):
            return "component/raspberry-mint/media.html.j2", {
                "file": document.resource,
            }
        return None


@ContentDefinition(
    "raspberry-mint-media-gallery",
    label=_("Media gallery"),
    description=_("Multiple files in a media gallery display"),
)
class MediaGallery(Template, Manufacturable):
    """
    Multiple files in a media gallery display.

    .. plugin:: content:raspberry-mint-media-gallery
    """

    @override
    @classmethod
    @require_project
    async def new(cls, project: Project, /) -> Self:
        await require_extension(RaspberryMint, project)
        return cls(jinja=await project.jinja)

    @override
    async def build_template(self, document: Document) -> TemplateBuild:
        if isinstance(document.resource, HasFileReferences):
            return "component/raspberry-mint/media-gallery.html.j2", {
                "file_references": list(associated_file_references(document.resource))
            }
        return None


@final
@ObjectDefinition(
    label=_("Color style configuration"),
    samples=[
        lambda: Sample(
            ColorStyleConfiguration(
                "my-first-content", style=RaspberryMintColorStyle.DARK
            ),
            label="Default",
        )
    ],
)
class ColorStyleConfiguration(Data):
    """
    Configuration for :py:class:`betty.extension.raspberry_mint.content.ColorStyle`.

    .. data:: betty.extension.raspberry_mint.content:ColorStyleConfiguration
    """

    content = PluginManufacturerSequenceProperty[ContentDefinition, Content](
        ContentManufacturer, label=_("Content")
    )
    """
    The content within this color style.
    """

    style = Property(EnumDefinition(cls=RaspberryMintColorStyle, label=_("Style")))
    """
    The style.
    """

    def __init__(
        self,
        content: ResolvablePluginManufacturerSequence[ContentDefinition, Content],
        *,
        style: RaspberryMintColorStyle,
    ):
        super().__init__()
        self.style = style
        self.content = content


@ContentDefinition("raspberry-mint-color-style", label=_("Color style"))
class ColorStyle(Template, DataManufacturable[ColorStyleConfiguration]):
    """
    Change the color style for all containing content.

    .. plugin:: content:raspberry-mint-color-style
    """

    def __init__(
        self,
        /,
        content: Iterable[Content],
        *,
        jinja: Environment,
        style: RaspberryMintColorStyle,
    ):
        super().__init__(jinja=jinja)
        self._content = tuple(content)
        self._style = style

    @override
    @classmethod
    def new_data_cls(cls) -> type[ColorStyleConfiguration]:
        return ColorStyleConfiguration

    @override
    @classmethod
    @require_project
    async def new(cls, project: Project, data: ColorStyleConfiguration, /) -> Self:
        await require_extension(RaspberryMint, project)
        content, jinja = await gather(
            gather(
                *map(
                    project.factory.new,
                    map(ContentManufacturer.resolve, data.content),
                )
            ),
            project.jinja,
        )
        return cls(content=content, jinja=jinja, style=data.style)

    @override
    async def build_template(self, document: Document) -> TemplateBuild:
        content = await build(document, self._content)
        if content is None:
            return None
        return "component/raspberry-mint/color-style.html.j2", {
            "color_style": self._style.value,
            "color_style_content": content,
        }


@ContentDefinition("raspberry-mint-external-links", label=_("External links"))
class ExternalLinks(Template, Manufacturable):
    """
    External links.

    .. plugin:: content:raspberry-mint-external-links
    """

    @override
    @classmethod
    @require_project
    async def new(cls, project: Project, /) -> Self:
        await require_extension(RaspberryMint, project)
        return cls(jinja=await project.jinja)

    @override
    async def build_template(self, document: Document) -> TemplateBuild:
        if isinstance(document.resource, HasLinks):
            return "component/raspberry-mint/links.html.j2", {
                "links": document.resource.links
            }
        return None


@ContentDefinition("raspberry-mint-timeline", label=_("Timeline"))
class Timeline(Template, Manufacturable):
    """
    A timeline of events.

    .. plugin:: content:raspberry-mint-timeline
    """

    def __init__(self, *, jinja: Environment, lifetime_threshold: int):
        super().__init__(jinja=jinja)
        self._lifetime_threshold = lifetime_threshold

    @override
    @classmethod
    @require_project
    async def new(cls, project: Project, /) -> Self:
        await require_extension(RaspberryMint, project)
        return cls(
            jinja=await project.jinja,
            lifetime_threshold=project.configuration.lifetime_threshold,
        )

    @override
    async def build_template(self, document: Document) -> TemplateBuild:
        events = []
        if isinstance(document.resource, Person):
            events.extend(
                person_timeline_events(document.resource, self._lifetime_threshold)
            )
        elif isinstance(document.resource, Place):
            events.extend(place_timeline_events(document.resource))
        if events:
            return "component/raspberry-mint/timeline.html.j2", {"events": events}
        return None


@ContentDefinition(
    "raspberry-mint-facts",
    label=_("Facts"),
    description=_(
        "Other entities that reference a citation or source to back up their claims."
    ),
)
class Facts(Template, Manufacturable):
    """
    A list of facts.

    .. plugin:: content:raspberry-mint-facts
    """

    @override
    @classmethod
    @require_project
    async def new(cls, project: Project, /) -> Self:
        await require_extension(RaspberryMint, project)
        return cls(jinja=await project.jinja)

    @override
    async def build_template(self, document: Document) -> TemplateBuild:
        entities = []
        if isinstance(document.resource, Citation):
            entities.extend(document.resource.facts)
        if isinstance(document.resource, Source):
            entities.extend(unique(self._source_facts(document.resource)))
        if entities:
            return "entity/list.html.j2", {"entities": entities}
        return None

    def _source_facts(self, source: Source) -> Iterable[Entity]:
        for citation in filter(is_public, source.citations):
            yield from citation.facts
        for contained in source.contains:
            yield from self._source_facts(contained)


@final
@ObjectDefinition(
    label=_("Presences configuration"),
    samples=[
        lambda: Sample(PresencesConfiguration(), label="Minimal"),
        lambda: Sample(
            PresencesConfiguration(include=["subject"]),
            label="Includes",
            size=Size.FULL,
        ),
        lambda: Sample(
            PresencesConfiguration(exclude=["subject"]),
            label="Excludes",
            size=Size.FULL,
        ),
    ],
)
class PresencesConfiguration(Data):
    """
    Configuration for :py:class:`betty.extension.raspberry_mint.content.Presences`.

    .. data:: betty.extension.raspberry_mint.content:PresencesConfiguration
    """

    exclude = Optional(
        Property(SequenceDefinition(cls=list, value=MachineName, label=_("Exclude")))
    )
    """
    The presence roles for which to exclude presences.
    """

    include = Optional(
        Property(SequenceDefinition(cls=list, value=MachineName, label=_("Include")))
    )
    """
    The presence roles for which to include presences.
    """

    def __init__(
        self,
        *,
        include: Iterable[ResolvableId[RoleDefinition]] | None = None,
        exclude: Iterable[ResolvableId[RoleDefinition]] | None = None,
    ):
        super().__init__()
        if include is not None:
            self.include = list(map(resolve_id, include))
        if exclude is not None:
            self.exclude = list(map(resolve_id, exclude))


@ContentDefinition("raspberry-mint-presences", label=_("Presences"))
class Presences(Template, DataManufacturable[PresencesConfiguration], Manufacturable):
    """
    People's presences at an event.

    .. plugin:: content:raspberry-mint-presences
    """

    def __init__(
        self,
        *,
        include: Iterable[ResolvableId[RoleDefinition]] | None = None,
        jinja: Environment,
    ):
        super().__init__(jinja=jinja)
        self._include = None if include is None else tuple(map(resolve_id, include))

    @override
    @classmethod
    def new_data_cls(cls) -> type[PresencesConfiguration]:
        return PresencesConfiguration

    @override
    @classmethod
    @require_project
    async def new(
        cls, project: Project, data: PresencesConfiguration | None = None, /
    ) -> Self:
        await require_extension(RaspberryMint, project)

        if data is None:
            raise NotImplementedError
        include: Iterable[ResolvableId[RoleDefinition]] | None
        if data.include is not None:
            include = data.include
        else:
            roles = await project.plugins.plugins(RoleDefinition)
            include = {role.id for role in roles}
            if data.exclude is not None:
                include -= set(data.exclude)
        return cls(include=include, jinja=await project.jinja)

    @override
    async def build_template(self, document: Document) -> TemplateBuild:
        if isinstance(document.resource, Event):
            presences = document.resource.presences
            if self._include is not None:
                presences = tuple(
                    presence
                    for presence in presences
                    if presence.role.plugin().id in self._include
                )
            if not len(presences):
                return None
            return "component/raspberry-mint/presences.html.j2", {
                "presences": presences
            }
        return None


type ColumnsWidth = Mapping[Breakpoint, Sequence[int]]
type ShorthandColumnsWidth = (
    int | Sequence[int] | Mapping[Breakpoint, int] | ColumnsWidth
)


@final
@ObjectDefinition(
    label=_("Columns configuration"),
    samples=[
        lambda: Sample(
            ColumnsConfiguration(
                [ContentManufacturer(Render, RenderConfiguration("Hello, world!"))]
            ),
            label="Minimal",
            size=Size.MINIMAL,
        ),
        lambda: Sample(
            ColumnsConfiguration(
                [ContentManufacturer(Render, RenderConfiguration("Hello, world!"))],
                justify_content=JustifyContent.CENTER,
            ),
            label="Justify content",
        ),
        lambda: Sample(
            ColumnsConfiguration(
                [ContentManufacturer(Render, RenderConfiguration("Hello, world!"))],
                width=6,
            ),
            label="A single column with a fixed, non-responsive width",
        ),
        lambda: Sample(
            ColumnsConfiguration(
                [
                    [
                        ContentManufacturer(
                            Render, RenderConfiguration("Hello, world!")
                        ),
                    ],
                    [
                        ContentManufacturer(
                            Render, RenderConfiguration("How are you?")
                        ),
                    ],
                ],
                width=[6, 6],
            ),
            label="Multiple columns with fixed, non-responsive widths",
        ),
        lambda: Sample(
            ColumnsConfiguration(
                [ContentManufacturer(Render, RenderConfiguration("Hello, world!"))],
                width={
                    Breakpoint.XS: 12,
                    Breakpoint.MD: 6,
                },
            ),
            label="A single column with responsive widths",
        ),
        lambda: Sample(
            ColumnsConfiguration(
                [
                    [
                        ContentManufacturer(
                            Render, RenderConfiguration("Hello, world!")
                        ),
                    ],
                    [
                        ContentManufacturer(
                            Render, RenderConfiguration("How are you?")
                        ),
                    ],
                ],
                width={
                    Breakpoint.XS: [12, 12],
                    Breakpoint.MD: [6, 6],
                },
            ),
            label="Multiple columns with responsive widths",
        ),
    ],
)
class ColumnsConfiguration(Data):
    """
    Configuration for :py:class:`betty.extension.raspberry_mint.content.Columns`.

    .. data:: betty.extension.raspberry_mint.content:ColumnsConfiguration
    """

    _DEFAULT_WIDTH: ColumnsWidth = {Breakpoint.XS: [12]}
    _width: ColumnsWidth

    content = Property(
        SequenceDefinition(
            cls=list,
            value=PluginManufacturerSequenceDefinition(
                ContentManufacturer, label=_("Column content")
            ),
            label=_("Columns"),
        )
    )
    """
    The content within the columns.
    """

    justify_content = Optional(
        Property(EnumDefinition(cls=JustifyContent, label=_("Justify content")))
    )
    """
    If and how to justify content.
    """

    width = Property(
        MappingDefinition(
            cls=dict,
            key=EnumDefinition(
                cls=Breakpoint,
                label=_("Breakpoint"),
            ),
            value=SequenceDefinition(
                cls=list,
                label=_("Column widths"),
                value=IntDefinition(label=_("Column width")),
            ),
            label=_("Breakpoints"),
            porter=CallbackPorter(
                assert_or(
                    assert_or(
                        assert_int(),
                        assert_sequence(assert_int()),
                    ),
                    assert_or(
                        assert_mapping(assert_int(), assert_enum(Breakpoint)),
                        assert_mapping(
                            assert_sequence(assert_int()), assert_enum(Breakpoint)
                        ),
                    ),
                ),
                lambda data: {
                    breakpoint.value: widths
                    for breakpoint, widths in data.items()  # noqa: A001
                },
            ),
        )
    )
    """
    The column widths.
    """

    def __init__(
        self,
        /,
        content: Sequence[
            ResolvablePluginManufacturerSequence[ContentDefinition, Content]
        ],
        *,
        width: ShorthandColumnsWidth | None = None,
        justify_content: JustifyContent | None = None,
    ):
        super().__init__()
        self.content = list(map(ContentManufacturer.resolve_sequence, content))
        if width is None:
            self._width = self._DEFAULT_WIDTH
        elif isinstance(width, int):
            self.width = {Breakpoint.XS: [width]}
        elif isinstance(width, Mapping):
            self.width = {
                breakpoint: [columns] if isinstance(columns, int) else columns
                for breakpoint, columns in width.items()  # noqa: A001
            }
        else:
            self.width = {Breakpoint.XS: width}
        self.justify_content = justify_content


@ContentDefinition("raspberry-mint-columns", label=_("Columns"))
class Columns(Template, DataManufacturable[ColumnsConfiguration]):
    """
    A container with one or more columns.

    .. plugin:: content:raspberry-mint-columns
    """

    def __init__(
        self,
        /,
        content: Iterable[Iterable[Content]],
        *,
        width: ColumnsWidth,
        jinja: Environment,
        justify_content: JustifyContent | None = None,
    ):
        super().__init__(jinja=jinja)
        self._content = tuple(map(tuple, content))
        self._justify_content = justify_content
        self._width = width

    @override
    @classmethod
    def new_data_cls(cls) -> type[ColumnsConfiguration]:
        return ColumnsConfiguration

    @override
    @classmethod
    @require_project
    async def new(cls, project: Project, data: ColumnsConfiguration, /) -> Self:
        await require_extension(RaspberryMint, project)
        content, jinja = await gather(
            gather(
                *[
                    gather(
                        *map(
                            project.factory.new,
                            map(ContentManufacturer.resolve, column_content),
                        )
                    )
                    for column_content in data.content
                ]
            ),
            project.jinja,
        )
        return cls(
            content=content,
            jinja=jinja,
            justify_content=data.justify_content,
            width=data.width,
        )

    @override
    async def build_template(self, document: Document) -> TemplateBuild:
        content = [
            await build(document, column_content) for column_content in self._content
        ]
        if not any(content):
            return None
        return "component/raspberry-mint/columns.html.j2", {
            "columns_content": content,
            "columns_justify_content": self._justify_content,
            "columns_width": {
                breakpoint.value: widths
                for breakpoint, widths in self._width.items()  # noqa: A001
            },
        }


@ContentDefinition("raspberry-mint-enclosees", label=_("Enclosees"))
class Enclosees(Template, Manufacturable):
    """
    Show the places enclosed by a place document resource.

    .. plugin:: content:raspberry-mint-enclosees
    """

    @override
    @classmethod
    @require_project
    async def new(cls, project: Project, /) -> Self:
        await require_extension(RaspberryMint, project)
        return cls(jinja=await project.jinja)

    @override
    async def build_template(self, document: Document) -> TemplateBuild:
        if isinstance(document.resource, Place):
            return "component/raspberry-mint/enclosees.html.j2", {
                "enclosees": list(self._enclosees(document.resource))
            }
        return None

    def _enclosees(self, place: Place) -> Iterable[Place]:
        for enclosure in place.enclosees:
            yield enclosure.enclosee
            yield from self._enclosees(enclosure.enclosee)


@ContentDefinition("raspberry-mint-file-referees", label=_("File referees"))
class FileReferees(Template, Manufacturable):
    """
    Show the entities referencing a document resource that is a file.

    .. plugin:: content:raspberry-mint-file-referees
    """

    @override
    @classmethod
    @require_project
    async def new(cls, project: Project, /) -> Self:
        await require_extension(RaspberryMint, project)
        return cls(jinja=await project.jinja)

    @override
    async def build_template(self, document: Document) -> TemplateBuild:
        if isinstance(document.resource, File):
            return "entity/list.html.j2", {
                "entities": [referee.referee for referee in document.resource.referees]
            }
        return None


@ContentDefinition("raspberry-mint-citations", label=_("Citations"))
class Citations(Template, Manufacturable):
    """
    The citations for a document resource that is an entity.

    .. plugin:: content:raspberry-mint-citations
    """

    @override
    @classmethod
    @require_project
    async def new(cls, project: Project, /) -> Self:
        await require_extension(RaspberryMint, project)
        return cls(jinja=await project.jinja)

    @override
    async def build_template(self, document: Document) -> TemplateBuild:
        if isinstance(document.resource, HasCitations):
            return "component/raspberry-mint/citations.html.j2", {
                "citations": document.resource.citations
            }
        return None



"""
Test utilities for :py:mod:`betty.job`.
"""

from typing import final, override

from betty.job import Job
from betty.job.executor.asyncio import AsyncExecutor
from betty.job.scheduler import Scheduler
from betty.job.scheduler.default import DefaultScheduler
from betty.user.no_op import NoOpUser


@final
class NoOpJob(Job):
    """
    A job that does nothing.
    """

    @override
    async def do(self, scheduler: Scheduler, /) -> None:
        pass


async def do(*jobs: Job) -> None:
    """
    Do a number of jobs.
    """
    scheduler = DefaultScheduler(user=NoOpUser())
    async with AsyncExecutor(scheduler):
        await scheduler.add(*jobs)
        async with scheduler:
            pass



"""
Interact with the Wikipedia Query API.
"""

from __future__ import annotations

from asyncio import to_thread
from contextlib import asynccontextmanager, contextmanager
from dataclasses import dataclass
from json import JSONDecodeError
from pathlib import Path
from typing import TYPE_CHECKING, Any, final
from urllib.parse import quote, urlparse

import aiofiles
from geopy import Point

from betty.assertion import assert_float, assert_mapping, assert_str
from betty.data.indicator import Url
from betty.data.indicator.selector import Index, Key, SelectorError, Selectors
from betty.exception import HumanFacingException, reraise_with_indicator
from betty.hashid import hashid
from betty.locale.localizable.gettext import _
from betty.media_type import MediaType

if TYPE_CHECKING:
    from collections.abc import AsyncIterator, Iterator, Mapping, MutableMapping

    from aiohttp import ClientResponse, ClientSession

    from betty.user import User


class ClientError(HumanFacingException, RuntimeError):
    """
    A client error.
    """


@final
@dataclass(frozen=True)
class Summary:
    """
    A Wikipedia page summary.
    """

    language: str
    name: str
    title: str
    content: str

    @property
    def url(self) -> str:
        """
        The URL to the web page.
        """
        return f"https://{self.language}.wikipedia.org/wiki/{self.name}"


@final
@dataclass(frozen=True)
class Image:
    """
    An image from Wikimedia Commons.
    """

    path: Path
    media_type: MediaType
    title: str
    wikimedia_commons_url: str
    name: str


class Client:
    """
    Fetch information from the Wikipedia Query API.
    """

    def __init__(
        self, *, download_directory_path: Path, http_client: ClientSession, user: User
    ):
        self._download_directory_path = download_directory_path
        self._http_client = http_client
        self._images: MutableMapping[str, Image | None] = {}
        self._user = user

    @contextmanager
    def _human_facing_exception_to_client_error(self) -> Iterator[None]:
        try:
            yield
        except HumanFacingException as error:
            raise ClientError(error) from error
        except SelectorError as error:
            raise ClientError(str(error)) from error

    @asynccontextmanager
    async def _get(self, url: str) -> AsyncIterator[ClientResponse]:
        async with self._http_client.get(url) as response:
            if response.status != 200:
                raise ClientError(
                    _("HTTP {http_status_code} response returned by {url}").format(
                        http_status_code=str(response.status), url=url
                    )
                )
            yield response

    async def _get_json(self, url: str) -> Any:
        async with self._get(url) as response:
            try:
                return await response.json()
            except JSONDecodeError as error:
                raise ClientError(f"Invalid JSON returned by {url}: {error}") from error

    async def _get_query_api_data(self, url: str) -> Mapping[str, Any]:
        data = await self._get_json(url)
        with (
            self._human_facing_exception_to_client_error(),
            reraise_with_indicator(Url(url)),
        ):
            return Selectors(Key("query"), Key("pages"), Index(0)).get(
                data, assert_mapping(None, assert_str())
            )

    async def _get_page_query_api_data(
        self, page_language: str, page_name: str
    ) -> tuple[str, Mapping[str, Any]]:
        url = f"https://{page_language}.wikipedia.org/w/api.php?action=query&prop=langlinks|pageimages|coordinates&lllimit=500&piprop=name&pilicense=free&pilimit=1&coprimary=primary&format=json&formatversion=2&titles={quote(page_name)}"
        return url, await self._get_query_api_data(url)

    async def get_translations(
        self, page_language: str, page_name: str
    ) -> Mapping[str, str]:
        """
        Get the available translations for a page.
        """
        _, api_data = await self._get_page_query_api_data(page_language, page_name)
        try:
            translations_data = api_data["langlinks"]
        except LookupError:
            # There may not be any translations.
            return {}
        return {
            translation_data["lang"]: translation_data["title"]
            for translation_data in translations_data
        }

    async def get_summary(self, page_language: str, page_name: str) -> Summary:
        """
        Get a summary for a page.
        """
        url = f"https://{page_language}.wikipedia.org/api/rest_v1/page/summary/{page_name}"

        with (
            self._human_facing_exception_to_client_error(),
            reraise_with_indicator(Url(url)),
        ):
            api_data = await self._get_json(url)

            title = Selectors(Key("titles"), Key("normalized")).get(
                api_data, assert_str()
            )
            extract = Selectors(
                Key("extract_html") if "extract_html" in api_data else Key("extract")
            ).get(api_data, assert_str())
        return Summary(
            page_language,
            page_name,
            title,
            extract,
        )

    async def get_image(self, page_language: str, page_name: str) -> Image | None:
        """
        Get an image for a page.
        """
        _, api_data = await self._get_page_query_api_data(page_language, page_name)
        try:
            page_image_name = api_data["pageimage"]
        except LookupError:
            # There may not be any images.
            return None

        if page_image_name in self._images:
            return self._images[page_image_name]

        url = f"https://en.wikipedia.org/w/api.php?action=query&prop=imageinfo&titles=File:{quote(page_image_name)}&iiprop=url|mime|canonicaltitle&format=json&formatversion=2"
        image_info_api_data = await self._get_query_api_data(url)

        image_info_selectors = (Key("imageinfo"), Index(0))
        with (
            self._human_facing_exception_to_client_error(),
            reraise_with_indicator(Url(url)),
        ):
            image_info = Selectors(*image_info_selectors).get(
                image_info_api_data, assert_mapping()
            )
            with reraise_with_indicator(*image_info_selectors):
                image_url = Key("url").get(image_info, assert_str())
                image_media_type = Key("mime").get(image_info, MediaType)
                image_title = Key("canonicaltitle").get(image_info, assert_str())
                image_wikimedia_commons_url = Key("descriptionurl").get(
                    image_info, assert_str()
                )
        async with self._get(image_url) as image_response:
            image_data = await image_response.read()
        image_path = (
            self._download_directory_path
            / "image"
            / (hashid(image_url) + Path(urlparse(image_url).path).suffix.lower())
        )
        await to_thread(image_path.parent.mkdir, exist_ok=True, parents=True)
        async with aiofiles.open(image_path, mode="wb") as image_f:
            await image_f.write(image_data)
        return Image(
            image_path,
            image_media_type,
            # Strip "File:" or any translated equivalent from the beginning of the image's title.
            image_title[image_title.index(":") + 1 :],
            image_wikimedia_commons_url,
            Path(urlparse(image_url).path).name,
        )

    async def get_place_coordinates(
        self, page_language: str, page_name: str
    ) -> Point | None:
        """
        Get the coordinates for a page that is a place.
        """
        url, api_data = await self._get_page_query_api_data(page_language, page_name)
        try:
            coordinates = api_data["coordinates"][0]
        except LookupError:
            # There may not be any coordinates.
            return None

        with (
            self._human_facing_exception_to_client_error(),
            reraise_with_indicator(Url(url)),
        ):
            globe = Key("globe").get(coordinates, assert_str())
            if globe != "earth":
                return None
            latitude = Key("lat").get(coordinates, assert_float())
            longitude = Key("lon").get(coordinates, assert_float())
        return Point(latitude, longitude)



"""
Object properties.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Self, cast, final, overload, override

from betty.data import OptionalDefinition
from betty.data.aggregate.record.object import Attr, AttrDefinition
from betty.functools import passthrough
from betty.importlib import fully_qualified_name
from betty.typing import Void

if TYPE_CHECKING:
    from collections.abc import Callable

    from betty.data import Data, DataDefinition
    from betty.locale.localizable import ResolvableLocalizable


class _Property[ValueGetT, ValueSetT](Attr[ValueGetT], ABC):
    _attr_name: str

    def __init__(
        self,
        attr: AttrDefinition[ValueGetT],
        *,
        resolver: Callable[[ValueSetT | ValueGetT], ValueGetT] = passthrough,
    ):
        self._attr = attr
        self._resolver = resolver

    def __set_name__(self, owner: type[Any], name: str) -> None:
        self._attr_name = f"_{name}"

    @override
    @property
    def attr(self) -> AttrDefinition[ValueGetT]:
        return self._attr

    @overload
    def __get__(self, instance: None, owner: type[object], /) -> Self:
        pass

    @overload
    def __get__(self, instance: Any, owner: type[Any], /) -> ValueGetT:
        pass

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return self.get(instance)

    def __set__(self, instance: Any, value: ValueSetT | ValueGetT) -> None:
        self.set(instance, value)

    @abstractmethod
    def get(self, instance: Any) -> ValueGetT:
        """
        Get the property value from the instance.
        """

    def set(self, instance: Any, value: ValueSetT | ValueGetT) -> ValueGetT:
        """
        Set the value on the instance.
        """
        resolved_value = self._resolver(value)
        setattr(instance, self._attr_name, resolved_value)
        return resolved_value


class Property[ValueGetT, ValueSetT](_Property[ValueGetT, ValueSetT]):
    """
    An object attribute with a definition.
    """

    def __init__(
        self,
        data: DataDefinition[ValueGetT] | type[Data[DataDefinition[ValueGetT]]],
        *,
        label: ResolvableLocalizable | None = None,
        description: ResolvableLocalizable | None = None,
        omit_load: bool | None = None,
        omit_dump: Callable[[ValueGetT], bool] | None = None,
        resolver: Callable[[ValueSetT | ValueGetT], ValueGetT] = passthrough,
        default: Callable[[], ValueGetT] | None = None,
    ):
        super().__init__(
            AttrDefinition(
                data,
                label=label,
                description=description,
                omit_load=omit_load,
                omit_dump=omit_dump,
            ),
            resolver=resolver,
        )
        self._data = data
        self._label = label
        self._description = description
        self._default = default

    @override
    def get(self, instance: Any) -> ValueGetT:
        value = cast(
            ValueGetT | Void,
            getattr(instance, self._attr_name, Void),
        )
        if value is Void:
            if self._default is None:
                instance_name = fully_qualified_name(type(instance))
                raise PropertyNotInitialized(
                    f"{instance_name}.{self._attr_name[1:]} was never initialized. Either provide a default when initializing the property, or make {instance_name}.__init__() set a value."
                )
            value = self._default()
            setattr(instance, self._attr_name, value)
        return value  # ty:ignore[invalid-return-type]


@final
class PropertyNotInitialized(ValueError):
    """
    Raised when a class failed to initialize a value for a :py:class:`betty.property.Property`.
    """


@final
class Optional[ValueGetT, ValueSetT](_Property[ValueGetT | None, ValueSetT | None]):
    """
    Make another property optional, e.g. allow ``None``.
    """

    def __init__(self, required_property: Property[ValueGetT, ValueSetT], /):
        def _omit_dump(data: ValueGetT | None) -> bool:
            if data is None:
                return True
            if required_property.attr.omit_dump is None:
                return False
            return required_property.attr.omit_dump(data)

        super().__init__(
            AttrDefinition(
                OptionalDefinition(required_property.attr.data),
                label=required_property.attr.label,
                description=required_property.attr.description,
                omit_load=required_property.attr.omit_load,
                omit_dump=_omit_dump,
            )
        )
        self._required_property = required_property

    def __set_name__(self, owner: type[Any], name: str) -> None:
        super().__set_name__(owner, name)
        self._required_property.__set_name__(owner, name)

    @override
    def get(self, instance: Any) -> ValueGetT | None:
        try:
            return self._required_property.get(instance)
        except PropertyNotInitialized:
            return self.set(instance, None)

    @override
    def set(self, instance: Any, value: ValueSetT | ValueGetT | None) -> ValueGetT:
        if value is None:
            return super().set(instance, value)
        return self._required_property.set(instance, value)

    def __delete__(self, instance: Any) -> None:
        self.delete(instance)

    def delete(self, instance: Any) -> None:
        """
        Delete the value from the instance.
        """
        self.set(instance, None)



"""
Provide proxy URL generators.
"""

from collections.abc import Mapping, Sequence
from typing import Any, final, override

from betty.locale import ResolvableLocale
from betty.media_type import MediaType
from betty.url import UnsupportedResource, UrlGenerator


@final
class ProxyUrlGenerator(UrlGenerator):
    """
    Expose multiple other URL generators as one unified URL generator.
    """

    def __init__(self, *upstreams: UrlGenerator):
        self._upstreams = upstreams

    @override
    def supports(self, resource: Any, /) -> bool:
        return any(upstream.supports(resource) for upstream in self._upstreams)

    @override
    def generate(
        self,
        resource: Any,
        *,
        absolute: bool = False,
        fragment: str | None = None,
        locale: ResolvableLocale | None = None,
        media_type: MediaType | None = None,
        query: Mapping[str, Sequence[str]] | None = None,
    ) -> str:
        for upstream in self._upstreams:
            if upstream.supports(resource):
                return upstream.generate(
                    resource,
                    absolute=absolute,
                    fragment=fragment,
                    locale=locale,
                    media_type=media_type,
                    query=query,
                )
        raise UnsupportedResource(resource)



"""
Provide the Generation API.
"""

from __future__ import annotations

import asyncio
import os
import shutil
from abc import ABC, abstractmethod
from asyncio import gather
from contextlib import suppress
from math import ceil
from os import cpu_count
from pathlib import Path
from typing import TYPE_CHECKING

from aiofiles.os import makedirs

from betty.concurrent import MAX_STRANDS
from betty.job import Context
from betty.job.executor.threading import ThreadPoolExecutor
from betty.job.scheduler.default import DefaultScheduler
from betty.project.generate.jobs import (
    GenerateEntitiesHtml,
    GenerateEntitiesJson,
    GenerateEntityTypesHtml,
    GenerateEntityTypesJson,
    GenerateFavicon,
    GenerateJsonErrorResponses,
    GenerateJsonSchema,
    GenerateLocalizedPublicAssets,
    GenerateOpenApi,
    GenerateRobotsTxt,
    GenerateSitemap,
    GenerateStaticPublicAssets,
)

if TYPE_CHECKING:
    from betty.job.scheduler import Scheduler
    from betty.project import Project


class Generator(ABC):
    """
    Generate (part of) a project's site.
    """

    @abstractmethod
    async def generate(self, scheduler: Scheduler) -> None:
        """
        Generate (part of) a project's site.
        """


async def generate(project: Project, *, context: Context | None = None) -> None:
    """
    Generate a new site.
    """
    if context is None:
        context = Context()

    await context.progress.add(2)

    await _preprocess(project)
    await context.progress.done()

    threading_concurrency = cpu_count() or 2
    scheduler = DefaultScheduler(context=context, user=project.app.user)
    async with ThreadPoolExecutor(
        scheduler,
        async_concurrency=ceil(MAX_STRANDS / threading_concurrency),
        threading_concurrency=threading_concurrency,
    ):
        await gather(
            *(
                extension.generate(scheduler)
                for extensions in await project.extensions
                for extension in extensions
                if isinstance(extension, Generator)
            )
        )
        await scheduler.release()
        await scheduler.add(
            GenerateStaticPublicAssets(project=project),
            GenerateSitemap(project=project),
            GenerateRobotsTxt(project=project),
            GenerateOpenApi(project=project),
            GenerateLocalizedPublicAssets(project=project),
            GenerateJsonSchema(project=project),
            GenerateJsonErrorResponses(project=project),
            GenerateFavicon(project=project),
            GenerateEntityTypesJson(project=project),
            GenerateEntityTypesHtml(project=project),
            GenerateEntitiesJson(project=project),
            GenerateEntitiesHtml(project=project),
        )
        await scheduler.complete()

    await _postprocess(project)
    await context.progress.done()


async def _preprocess(project: Project) -> None:
    await _preprocess_output_directory(project.output_directory)
    await _preprocess_www_directory(project.www_directory)


async def _preprocess_output_directory(output_directory_path: Path) -> None:
    with suppress(FileNotFoundError):
        await asyncio.to_thread(shutil.rmtree, output_directory_path)
    await makedirs(output_directory_path, exist_ok=True)


async def _preprocess_www_directory(www_directory_path: Path) -> None:
    await makedirs(www_directory_path, exist_ok=True)


async def _postprocess(project: Project) -> None:
    await _postprocess_output_directory(project.output_directory)


async def _postprocess_output_directory(output_directory_path: Path) -> None:
    output_directory_path.chmod(0o755)
    for directory_path_str, subdirectory_names, file_names in os.walk(
        output_directory_path
    ):
        directory_path = Path(directory_path_str)
        for subdirectory_name in subdirectory_names:
            (directory_path / subdirectory_name).chmod(0o755)
        for file_name in file_names:
            (directory_path / file_name).chmod(0o644)



"""
Provide Betty's default Jinja2 tests.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from betty.ancestry.has_file_references import HasFileReferences
from betty.image import is_supported_media_type
from betty.json.linked_data import LinkedDataDumpableWithSchema
from betty.model import persistent_id
from betty.plugin import Plugin, PluginDefinition
from betty.privacy import is_public
from betty.service.level import UNIVERSE
from betty.string import kebab_case_to_snake_case

if TYPE_CHECKING:
    from collections.abc import Callable, Mapping, MutableMapping

    from betty.machine_name import MachineName
    from betty.media_type import MediaType


def test_linked_data_dumpable(value: Any) -> bool:
    """
    Test if a value can be dumped to Linked Data.
    """
    return isinstance(value, LinkedDataDumpableWithSchema)


class PluginTester:
    """
    Provides tests for a specific plugin type.
    """

    def __init__(self, plugin_type: type[PluginDefinition], /):
        self._plugin_type = plugin_type

    def tests(self) -> Mapping[str, Callable[..., bool]]:
        """
        Get the available tests, keyed by test name.
        """
        return {f"{kebab_case_to_snake_case(self._plugin_type.type().id)}_plugin": self}

    def __call__(self, /, value: Any, plugin_id: MachineName | None = None) -> bool:
        """
        :param plugin_id: If given, additionally ensure the value is an instance of this type.
        """
        if not isinstance(value, Plugin):
            return False
        if not isinstance(value.plugin(), self._plugin_type):
            return False
        return not (plugin_id is not None and value.plugin().id != plugin_id)


def test_has_file_references(value: Any) -> bool:
    """
    Test if a value has :py:class:`betty.ancestry.file_reference.FileReference` entities associated with it.
    """
    return isinstance(value, HasFileReferences)


def test_image_supported_media_type(media_type: MediaType | None) -> bool:
    """
    Test if a media type is supported by the image API.
    """
    if media_type is None:
        return False
    return is_supported_media_type(media_type)


async def tests() -> Mapping[str, Callable[..., bool]]:
    """
    Define the available tests.
    """
    tests: MutableMapping[str, Callable[..., bool]] = {
        "has_file_references": test_has_file_references,
        "persistent_entity_id": persistent_id,
        "image_supported_media_type": test_image_supported_media_type,
        "linked_data_dumpable": test_linked_data_dumpable,
        "public": is_public,
    }
    for plugin in UNIVERSE.plugins.types:
        tests.update(PluginTester(plugin).tests())
    return tests



"""
JSON schemas for the media type API.
"""

from __future__ import annotations

from typing import final

from betty.classtools import Singleton
from betty.json.schema import String


@final
class MediaTypeSchema(Singleton, String):
    """
    A JSON Schema for :py:class:`betty.media_type.MediaType`.
    """

    def __init__(self):
        super().__init__(
            def_name="mediaType",
            title="Media type",
            description="An IANA media type (https://www.iana.org/assignments/media-types/media-types.xhtml).",
        )



"""
Data for the Raspberry Mint extension.
"""

from __future__ import annotations

from collections.abc import Iterable, Mapping
from typing import TYPE_CHECKING, final

from betty.color import ColorDefinition
from betty.content import (
    Content,
    ContentDefinition,
    ContentManufacturer,
)
from betty.data import Data, Sample
from betty.data.aggregate.collection.mapping import MappingDefinition
from betty.data.aggregate.record.object import ObjectDefinition
from betty.data.indicator.selector import Attr, Key
from betty.data.str import StrDefinition
from betty.exception import HumanFacingException, reraise_with_indicator
from betty.locale.localizable.gettext import _
from betty.locale.localizable.markup import Paragraph, do_you_mean
from betty.plugin.data import PluginManufacturerSequenceDefinition
from betty.plugin.factory import ResolvablePluginManufacturer
from betty.property import Property
from betty.sample import Size
from betty.service.requirement.extension import require_extension

if TYPE_CHECKING:
    from betty.extension.raspberry_mint import RaspberryMint

type ResolvableRegionalContent = Mapping[
    str,
    Iterable[ResolvablePluginManufacturer[ContentDefinition, Content]],
]


@final
@ObjectDefinition(
    label=_("Raspberry Mint configuration"),
    samples=[
        lambda: Sample(
            RaspberryMintConfiguration(), label="Minimal", size=Size.MINIMAL
        ),
        lambda: Sample(
            RaspberryMintConfiguration(
                primary_color=ColorDefinition().samples.get(Size.MINIMAL).subject,
                secondary_color=ColorDefinition().samples.get(Size.MINIMAL).subject,
                tertiary_color=ColorDefinition().samples.get(Size.MINIMAL).subject,
            ),
            label="Custom colors",
        ),
        lambda: Sample(
            RaspberryMintConfiguration(
                regional_content={
                    "front-page-content": {
                        "id": "render",
                        "configuration": {
                            "content": "Hello, world!",
                        },
                    }
                }
            ),
            label="Regional content",
        ),
    ],
)
class RaspberryMintConfiguration(Data):
    """
    Configuration for the :py:class:`betty.extension.raspberry_mint.RaspberryMint` extension.

    .. data:: betty.extension.raspberry_mint.data:RaspberryMintConfiguration
    """

    DEFAULT_PRIMARY_COLOR = "#b3446c"
    DEFAULT_SECONDARY_COLOR = "#3eb489"
    DEFAULT_TERTIARY_COLOR = "#ffbd22"

    primary_color = Property(
        ColorDefinition(),
        label=_("Primary color"),
        default=lambda: RaspberryMintConfiguration.DEFAULT_PRIMARY_COLOR,
    )
    """
    The primary color.
    """

    secondary_color = Property(
        ColorDefinition(),
        label=_("Secondary color"),
        default=lambda: RaspberryMintConfiguration.DEFAULT_SECONDARY_COLOR,
    )
    """
    The secondary color.
    """

    tertiary_color = Property(
        ColorDefinition(),
        label=_("Tertiary color"),
        default=lambda: RaspberryMintConfiguration.DEFAULT_TERTIARY_COLOR,
    )
    """
    The tertiary color.
    """

    regional_content = Property(
        MappingDefinition(
            cls=dict,
            label=_("Regions"),
            key=StrDefinition(label=_("Region")),
            value=PluginManufacturerSequenceDefinition(
                ContentManufacturer, label=_("Regional content")
            ),
        ),
        default=dict,
        omit_load=True,
        omit_dump=lambda data: not len(data),
    )
    """
    The regional content.
    """

    def __init__(
        self,
        *,
        primary_color: str | None = None,
        secondary_color: str | None = None,
        tertiary_color: str | None = None,
        regional_content: ResolvableRegionalContent | None = None,
    ):
        super().__init__()
        if primary_color is not None:
            self.primary_color = primary_color
        if secondary_color is not None:
            self.secondary_color = secondary_color
        if tertiary_color is not None:
            self.tertiary_color = tertiary_color
        if regional_content is not None:
            self.regional_content.update(
                {
                    region: ContentManufacturer.resolve_sequence(content)
                    for region, content in regional_content.items()
                }
            )

    @require_extension("raspberry-mint")
    async def validate(self, raspberry_mint: RaspberryMint, /) -> None:
        """
        Validate the configuration.
        """
        available_regions = await raspberry_mint.regions
        with reraise_with_indicator(Attr("regional_content")):
            for region in self.regional_content:
                with reraise_with_indicator(Key(region)):
                    if region not in available_regions:
                        raise HumanFacingException(
                            Paragraph(
                                _("Invalid region {invalid_region}.").format(
                                    invalid_region=f'"{region}"',
                                ),
                                do_you_mean(
                                    *(
                                        f'"{available_region}"'
                                        for available_region in available_regions
                                    )
                                ),
                            )
                        ) from None



"""
Provide exception handling utilities.
"""

from __future__ import annotations

from contextlib import contextmanager
from typing import TYPE_CHECKING, Never, override

from betty.data.indicator.selector import Selectors
from betty.locale.localizable import Localizable, ResolvableLocalizable
from betty.locale.localize import resolve_localized

if TYPE_CHECKING:
    from collections.abc import Iterator, Sequence

    from ty_extensions import Intersection

    from betty.data.indicator import Indicator
    from betty.locale import HasLocale
    from betty.locale.localize import Localizer


def do_raise(exception: BaseException, /) -> Never:
    """
    Raise the given exception.

    This is helpful as a callback.
    """
    raise exception


@contextmanager
def reraise_with_indicator(*indicators: Indicator) -> Iterator[None]:
    """
    Re-raise a human-facing exception with the given indicators.
    """
    try:
        yield
    except HumanFacingException as error:
        error.with_indicator(*indicators)
        raise


class HumanFacingException(Exception, Localizable):
    """
    A localizable, human-facing exception.

    When encountering an exception that extends this base class, Betty will show the localized exception message, and
    no stack trace.
    """

    def __init__(
        self,
        message: ResolvableLocalizable,
        *,
        indicators: Sequence[Indicator] | None = None,
    ):
        from betty.locale.localize import DEFAULT_LOCALIZER

        super().__init__(
            # Provide a default localization so this exception can be displayed like any other.
            resolve_localized(message, localizer=DEFAULT_LOCALIZER),
        )
        self._localizable_message = message
        self._indicators = [] if indicators is None else list(indicators)

    @override
    def __str__(self) -> str:
        from betty.locale.localize import DEFAULT_LOCALIZER

        return self.localize(DEFAULT_LOCALIZER)

    @override
    def localize(self, localizer: Localizer, /) -> Intersection[HasLocale, str]:
        from betty.locale.localizable.markup import Lines, UnorderedList

        return Lines(
            self._localizable_message,
            UnorderedList(
                *[
                    selector.format()
                    for selector in Selectors.reduce(*reversed(self.indicators))
                ]
            ),
        ).localize(localizer)

    @property
    def indicators(self) -> Sequence[Indicator]:
        """
        Get the human-readable indicators describing where the error occurred in the source data.

        The first indicator is the innermost, and the last indicator is the outermost.
        """
        return self._indicators

    def with_indicator(self, *indicators: Indicator) -> None:
        """
        Adds the given indicator(s) to the exception.

        The first indicator is the innermost, and the last indicator is the outermost.
        """
        self._indicators.extend(indicators)



"""
Object data types.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections import defaultdict
from inspect import getmembers
from types import FunctionType
from typing import TYPE_CHECKING, final, override

from betty.data import DataDefinition
from betty.data.aggregate.record import FieldDefinition, RecordDefinition
from betty.data.indicator.selector import Attr as AttrElement
from betty.importlib import fully_qualified_name
from betty.locale.localizable import resolve_localizable

if TYPE_CHECKING:
    from collections.abc import Callable, MutableMapping

    from betty.data import Data
    from betty.locale.localizable import Localizable, ResolvableLocalizable


_attrs: MutableMapping[str, MutableMapping[str, AttrDefinition]] = defaultdict(dict)


@final
class AttrDefinition[DataClsT]:
    """
    Define an object attribute.

    Usage:

    .. code-block:: python

       @ObjectDefinition(label="My first object")
       class MyFirstObject(Data[ObjectDefinition]):
           @property
           @AttrDefinition(BoolDefinition(label="My first attribute"))
           def my_first_attr(self) -> bool:
               return True
    """

    def __init__(
        self,
        data: DataDefinition[DataClsT] | type[Data[DataDefinition[DataClsT]]],
        *,
        label: ResolvableLocalizable | None = None,
        description: ResolvableLocalizable | None = None,
        omit_load: bool | None = None,
        omit_dump: Callable[[DataClsT], bool] | None = None,
    ):
        self._data = data if isinstance(data, DataDefinition) else data.data()
        self._label = None if label is None else resolve_localizable(label)
        self._description = (
            None if description is None else resolve_localizable(description)
        )
        self._omit_load = omit_load
        self._omit_dump = omit_dump

    def field(self, name: str, /) -> FieldDefinition:
        """
        Create a field definition for this attribute.
        """
        return FieldDefinition(
            AttrElement(name),
            self._data,
            label=self._label,
            description=self._description,
            omit_load=self._omit_load,
            omit_dump=self._omit_dump,
        )

    def __call__[FunctionTypeT: FunctionType](
        self, attribute: FunctionTypeT
    ) -> FunctionTypeT:
        """
        Decorate an attribute.
        """
        global _attrs
        cls_name, attribute_name = (
            f"{attribute.__globals__['__spec__'].name}:{attribute.__qualname__}".rsplit(
                ".", 1
            )
        )
        _attrs[cls_name][attribute_name] = self
        return attribute

    @property
    def data(self) -> DataDefinition[DataClsT]:
        """
        The attribute's data definition.
        """
        return self._data

    @property
    def label(self) -> Localizable | None:
        """
        The human-readable attribute label.
        """
        return self._label

    @property
    def description(self) -> Localizable | None:
        """
        The human-readable long attribute description.
        """
        return self._description

    @property
    def omit_load(self) -> bool | None:
        """
        Check if the field may be omitted from the parent when loading from portable data.
        """
        return self._omit_load

    @property
    def omit_dump(self) -> Callable[[DataClsT], bool] | None:
        """
        Check if the field may be omitted from the parent when dumping to portable data.
        """
        return self._omit_dump


class Attr[DataClsT](ABC):
    """
    A class attribute that exposes its data definition.
    """

    @property
    @abstractmethod
    def attr(self) -> AttrDefinition[DataClsT]:
        """
        The attribute's data definition.
        """


class ObjectDefinition[DataClsT](RecordDefinition[DataClsT, AttrElement]):
    """
    Define an object with attributes.

    Use :py:class:`betty.data.aggregate.record.object.AttrDefinition` to define fields inline, or in superclasses so they
    can be inherited.
    """

    @override
    def _set_cls(self, cls: type[DataClsT]) -> None:
        global _attrs
        super()._set_cls(cls)
        cls_name = fully_qualified_name(cls)
        attrs = _attrs[cls_name]

        for member_name, member in getmembers(cls):
            if isinstance(member, Attr):
                self._fields.append(member.attr.field(member_name))
            elif member_name in attrs:
                self._fields.append(
                    attrs[member_name].field(member_name),  # ty:ignore[invalid-argument-type]
                )
        return cls  # ty:ignore[invalid-return-type]



"""
Data types to represent names.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, final, override

from betty.ancestry.date import HasDate
from betty.locale.localizable.gettext import _
from betty.locale.localizable.linked_data import dump_linked_data
from betty.locale.localizable.property import LocalizableProperty
from betty.locale.localizable.static.schema import StaticTranslationsSchema

if TYPE_CHECKING:
    from betty.date import ResolvableDate
    from betty.json.linked_data import JsonLdObject
    from betty.locale.localizable import ResolvableLocalizable
    from betty.portable import PortableMapping
    from betty.project import Project


@final
class Name(HasDate):
    """
    A name.

    A name can be translated, and have a date expressing the period the name was in use.
    """

    name = LocalizableProperty(label=_("Name"))

    def __init__(
        self,
        name: ResolvableLocalizable,
        *,
        date: ResolvableDate | None = None,
    ):
        super().__init__(date=date)
        self.name = name

    @override
    @classmethod
    async def linked_data_schema(cls, project: Project, /) -> JsonLdObject:
        schema = await super().linked_data_schema(project)
        schema.add_property("name", StaticTranslationsSchema())
        return schema

    @override
    async def dump_linked_data(self, project: Project, /) -> PortableMapping:
        portable = await super().dump_linked_data(project)
        portable["name"] = dump_linked_data(
            self.name, localizers=await project.public_localizers
        )
        return portable



"""
Test utilities for :py:mod:`betty.plugin.manager`.
"""

from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING, cast, override

from betty.plugin import (
    Plugin,
    PluginDefinition,
    PluginTypeRepository,
    ResolvableDefinition,
    resolve_definition,
)
from betty.plugin.manager import PluginManager
from betty.plugin.repository.static import StaticPluginRepository
from betty.typing import threadsafe

if TYPE_CHECKING:
    from collections.abc import Collection, Iterable, Mapping

    from betty.machine_name import ResolvableMachineName
    from betty.plugin.repository import PluginRepository


@threadsafe
class StaticPluginManager(PluginManager):
    """
    Manage static plugin definitions.
    """

    def __init__(
        self,
        plugins: Mapping[
            type[PluginDefinition],
            ResolvableDefinition | Iterable[ResolvableDefinition],
        ],
        /,
    ):
        self._plugins: Mapping[type[PluginDefinition], Collection[PluginDefinition]] = (
            defaultdict(
                tuple,
                {
                    plugin_type: [resolve_definition(plugin_type_plugins)]
                    if isinstance(plugin_type_plugins, PluginDefinition)
                    else [plugin_type_plugins]
                    if isinstance(plugin_type_plugins, type)
                    and issubclass(plugin_type_plugins, Plugin)
                    else plugin_type_plugins
                    for plugin_type, plugin_type_plugins in plugins.items()
                },
            )  # ty:ignore[invalid-assignment]
        )
        self._types = PluginTypeRepository()

    @override
    async def plugins[PluginDefinitionT: PluginDefinition](
        self, plugin_type: type[PluginDefinitionT] | ResolvableMachineName, /
    ) -> PluginRepository[PluginDefinitionT]:
        if isinstance(plugin_type, str):
            plugin_type = cast(type[PluginDefinitionT], self.types[plugin_type])
        return StaticPluginRepository(plugin_type, *self._plugins[plugin_type])  # ty:ignore[invalid-return-type]

    @override
    @property
    def types(self) -> PluginTypeRepository:
        return self._types



"""
Content plugins.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, final, override

from markupsafe import Markup

from betty.definition.human_facing import HumanFacingDefinition
from betty.locale.localizable.gettext import _, ngettext
from betty.plugin import Plugin, PluginDefinition, PluginTypeDefinition
from betty.plugin.discovery.entry_point import EntryPointDiscovery
from betty.plugin.factory import PluginManufacturer

if TYPE_CHECKING:
    import builtins
    from collections.abc import Iterable

    from betty.document import Document


class Content(ABC, Plugin["ContentDefinition"]):
    """
    A content plugin.
    """

    @abstractmethod
    async def build(self, *, document: Document) -> str | None:
        """
        Render the content.
        """


@final
@PluginTypeDefinition(
    "content",
    label=_("Content"),
    label_plural=_("Contents"),
    label_countable=ngettext("{count} content", "{count} contents"),
    discovery=[EntryPointDiscovery("betty.content")],
)
class ContentDefinition(HumanFacingDefinition, PluginDefinition[Content]):
    """
    .. plugin_type:: content.
    """


@final
class ContentManufacturer(PluginManufacturer[ContentDefinition, Content]):
    """
    The content manufacturer.
    """

    @override
    @classmethod
    def type(cls) -> builtins.type[ContentDefinition]:
        return ContentDefinition


async def build(document: Document, contents: Iterable[Content], /) -> Markup | None:
    """
    Build content for the given document and contents.
    """
    built = "".join(
        [await content.build(document=document) or "" for content in contents]
    ).strip()
    if built:
        return Markup(built)
    return None



"""
Jobs.
"""

from __future__ import annotations

from asyncio import gather, to_thread
from io import BytesIO
from json import dumps
from math import ceil
from pathlib import Path
from typing import TYPE_CHECKING, cast, final, override

import aiofiles
from aiofiles.os import makedirs
from PIL import Image

from betty.job import Job
from betty.locale.localizable.gettext import _
from betty.locale.localize import DEFAULT_LOCALIZER
from betty.media_type.media_types import HTML, JSON
from betty.model import EntityDefinition, persistent_id
from betty.openapi import Specification
from betty.privacy import is_public
from betty.project.generate.file import (
    create_file,
    create_html_resource,
    create_json_resource,
)
from betty.project.schema import ProjectSchema
from betty.string import kebab_case_to_lower_camel_case

if TYPE_CHECKING:
    from collections.abc import MutableSequence

    from babel import Locale

    from betty.jinja import CopyFunction
    from betty.job.scheduler import Scheduler
    from betty.portable import PortableMapping
    from betty.project import Project


@final
class GenerateStaticPublicAssets(Job):
    """
    Generate a site's static public assets.
    """

    def __init__(self, *, project: Project):
        super().__init__(self.id_for(), priority=True)
        self._project = project

    @classmethod
    def id_for(cls) -> str:
        """
        Get the job ID.
        """
        return "generate-static-public-assets"

    @override
    async def do(self, scheduler: Scheduler, /) -> None:

        assets = await self._project.assets
        jinja = await self._project.jinja
        copy_function = jinja.make_copy_function(
            document=await self._project.new_document(context=scheduler.context),
            www_directory_path=self._project.www_directory,
            is_localized_and_multilingual=self._project.configuration.multilingual,
        )
        await gather(
            *[
                self._generate(scheduler, asset_path, copy_function)
                async for asset_path in assets.walk(Path("public") / "static")
            ]
        )

    async def _generate(
        self,
        scheduler: Scheduler,
        asset_path: Path,
        copy_function: CopyFunction,
    ) -> None:

        assets = await self._project.assets
        file_destination_path = self._project.www_directory / asset_path.relative_to(
            Path("public") / "static"
        )
        await makedirs(file_destination_path.parent, exist_ok=True)
        await copy_function(await assets.get(asset_path), file_destination_path)


@final
class GenerateSitemap(Job):
    """
    Generate a site's sitemap.
    """

    _SITEMAP_URL_TEMPLATE = """<url>
        <loc>{{{ loc }}}</loc>
        <lastmod>{{{ lastmod }}}</lastmod>
    </url>
    """

    _SITEMAP_BATCH_TEMPLATE = """<?xml version="1.0" encoding="utf-8"?>
    <urlset xmlns="http://www.sitemaps.org/schemas/sitemap/0.9" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://www.sitemaps.org/schemas/sitemap/0.9 http://www.sitemaps.org/schemas/sitemap/0.9/sitemap.xsd">
        {{{ urls }}}
    </urlset>
    """

    _SITEMAP_SITEMAP_TEMPLATE = """<sitemap>
        <loc>{{{ loc }}}</loc>
    </sitemap>
    """

    _SITEMAP_TEMPLATE = """<?xml version="1.0" encoding="UTF-8"?>
    <sitemapindex xmlns="http://www.sitemaps.org/schemas/sitemap/0.9">
        {{{ sitemaps }}}
    </sitemapindex>
    """

    def __init__(self, *, project: Project):
        super().__init__(self.id_for())
        self._project = project

    @classmethod
    def id_for(cls) -> str:
        """
        Get the job ID.
        """
        return "generate-sitemap"

    @override
    async def do(self, scheduler: Scheduler, /) -> None:
        context = scheduler.context

        url_generator = await self._project.url_generator

        await to_thread(
            self._project.www_directory.mkdir,
            exist_ok=True,
            parents=True,
        )

        sitemap_batches = []
        sitemap_batch_urls: MutableSequence[str] = []
        sitemap_batch_urls_length = 0
        sitemap_batches.append(sitemap_batch_urls)
        for locale in self._project.configuration.locales.keys():  # noqa: SIM118
            for entity in self._project.ancestry:
                if not persistent_id(entity):
                    continue
                if not entity.plugin().public_facing:
                    continue

                sitemap_batch_urls.append(
                    url_generator.generate(
                        entity,
                        absolute=True,
                        locale=locale,
                        media_type=HTML,
                    )
                )
                sitemap_batch_urls_length += 1

                if sitemap_batch_urls_length == 50_000:
                    sitemap_batch_urls = []
                    sitemap_batch_urls_length = 0
                    sitemap_batches.append(sitemap_batch_urls)

        sitemap_urls = []
        for sitemap_batch_index, sitemap_batch_urls in enumerate(sitemap_batches):
            sitemap_urls.append(
                url_generator.generate(
                    f"betty-static:///sitemap-{sitemap_batch_index}.xml",
                    absolute=True,
                )
            )
            rendered_sitemap_batch = self._SITEMAP_BATCH_TEMPLATE.replace(
                "{{{ urls }}}",
                "".join(
                    self._SITEMAP_URL_TEMPLATE.replace(
                        "{{{ loc }}}", sitemap_batch_url
                    ).replace("{{{ lastmod }}}", context.start.isoformat())
                    for sitemap_batch_url in sitemap_batch_urls
                ),
            )
            async with aiofiles.open(
                self._project.www_directory / f"sitemap-{sitemap_batch_index}.xml",
                "w",
            ) as f:
                await f.write(rendered_sitemap_batch)

        rendered_sitemap = self._SITEMAP_TEMPLATE.replace(
            "{{{ sitemaps }}}",
            "".join(
                self._SITEMAP_SITEMAP_TEMPLATE.replace("{{{ loc }}}", sitemap_url)
                for sitemap_url in sitemap_urls
            ),
        )
        async with aiofiles.open(self._project.www_directory / "sitemap.xml", "w") as f:
            await f.write(rendered_sitemap)


@final
class GenerateRobotsTxt(Job):
    """
    Generate a site's robots.txt.
    """

    _ROBOTS_TXT_TEMPLATE = """Sitemap: {{{ sitemap }}}"""

    def __init__(self, *, project: Project):
        super().__init__(self.id_for())
        self._project = project

    @classmethod
    def id_for(cls) -> str:
        """
        Get the job ID.
        """
        return "generate-robots-txt"

    @override
    async def do(self, scheduler: Scheduler, /) -> None:

        url_generator = await self._project.url_generator
        rendered_robots_txt = self._ROBOTS_TXT_TEMPLATE.replace(
            "{{{ sitemap }}}",
            url_generator.generate("betty-static:///sitemap.xml", absolute=True),
        )
        await to_thread(
            self._project.www_directory.mkdir,
            exist_ok=True,
            parents=True,
        )
        async with aiofiles.open(
            self._project.www_directory / "robots.txt", mode="w"
        ) as f:
            await f.write(rendered_robots_txt)


@final
class GenerateOpenApi(Job):
    """
    Generate a site's OpenAPI specification.
    """

    def __init__(self, *, project: Project):
        super().__init__(self.id_for())
        self._project = project

    @classmethod
    def id_for(cls) -> str:
        """
        Get the job ID.
        """
        return "generate-openapi"

    @override
    async def do(self, scheduler: Scheduler, /) -> None:

        api_directory_path = self._project.www_directory / "api"
        rendered_json = dumps(await Specification(self._project).build())
        async with create_json_resource(api_directory_path) as f:
            await f.write(rendered_json)


@final
class GenerateLocalizedPublicAssets(Job):
    """
    Generate a site's localized public assets.
    """

    def __init__(self, *, project: Project):
        super().__init__(
            self.id_for(),
            dependencies={GenerateStaticPublicAssets.id_for()},
            priority=True,
        )
        self._project = project

    @classmethod
    def id_for(cls) -> str:
        """
        Get the job ID.
        """
        return "generate-localized-public-assets"

    @override
    async def do(self, scheduler: Scheduler, /) -> None:

        assets = await self._project.assets
        localizers = await self._project.localizers
        jinja = await self._project.jinja
        copy_functions = {
            locale: jinja.make_copy_function(
                document=await self._project.new_document(
                    context=scheduler.context,
                    localizer=localizers.get(locale),
                ),
                www_directory_path=self._project.www_directory,
                is_localized_and_multilingual=self._project.configuration.multilingual,
            )
            for locale in self._project.configuration.locales.keys()  # noqa: SIM118
        }
        await gather(
            *[
                self._generate(scheduler, asset_path, copy_functions[locale], locale)
                async for asset_path in assets.walk(Path("public") / "localized")
                for locale in self._project.configuration.locales.keys()  # noqa: SIM118
            ]
        )

    async def _generate(
        self,
        scheduler: Scheduler,
        asset_path: Path,
        copy_function: CopyFunction,
        locale: Locale,
    ) -> None:

        assets = await self._project.assets
        file_destination_path = self._project.localize_www_directory(
            locale
        ) / asset_path.relative_to(Path("public") / "localized")
        await makedirs(file_destination_path.parent, exist_ok=True)
        await copy_function(await assets.get(asset_path), file_destination_path)


@final
class GenerateJsonSchema(Job):
    """
    Generate the JSON schema for a site.
    """

    def __init__(self, *, project: Project):
        super().__init__(self.id_for())
        self._project = project

    @classmethod
    def id_for(cls) -> str:
        """
        Get the job ID.
        """
        return "generate-json-schema"

    @override
    async def do(self, scheduler: Scheduler, /) -> None:

        schema = await ProjectSchema.new(self._project)
        rendered_json = dumps(schema.schema)
        async with create_file(ProjectSchema.www_path(self._project)) as f:
            await f.write(rendered_json)


@final
class GenerateJsonErrorResponses(Job):
    """
    Generate JSON HTTP error responses.
    """

    def __init__(self, *, project: Project):
        super().__init__(self.id_for())
        self._project = project

    @classmethod
    def id_for(cls) -> str:
        """
        Get the job ID.
        """
        return "generate-json-error-responses"

    @override
    async def do(self, scheduler: Scheduler, /) -> None:

        for code, message in [
            (401, _("I'm sorry, dear, but it seems you're not logged in.")),
            (
                403,
                _(
                    "I'm sorry, dear, but it seems you're not allowed to view this page."
                ),
            ),
            (404, _("I'm sorry, dear, but it seems this page does not exist.")),
        ]:
            for locale in self._project.configuration.locales.keys():  # noqa: SIM118
                async with create_file(
                    self._project.localize_www_directory(locale)
                    / ".error"
                    / f"{code}.json"
                ) as f:
                    await f.write(
                        dumps(
                            {
                                "$schema": await ProjectSchema.def_url(
                                    self._project, "errorResponse"
                                ),
                                "message": message.localize(DEFAULT_LOCALIZER),
                            }
                        )
                    )


@final
class GenerateFavicon(Job):
    """
    Generate a site's favicon.
    """

    def __init__(self, *, project: Project):
        super().__init__(self.id_for())
        self._project = project

    @classmethod
    def id_for(cls) -> str:
        """
        Get the job ID.
        """
        return "generate-favicon"

    @override
    async def do(self, scheduler: Scheduler, /) -> None:

        await to_thread(
            self._project.www_directory.mkdir,
            exist_ok=True,
            parents=True,
        )

        async with aiofiles.open(self._project.logo, "rb") as logo_f:
            logo = BytesIO(await logo_f.read())
        image = Image.open(logo)
        favicon = BytesIO()
        image.save(favicon, format="ICO")
        async with aiofiles.open(
            self._project.www_directory / "favicon.ico", "wb"
        ) as favicon_f:
            await favicon_f.write(favicon.getbuffer())


@final
class GenerateEntityTypesJson(Job):
    """
    Generate JSON resources for entity types.
    """

    def __init__(self, *, project: Project):
        super().__init__(self.id_for(), priority=True)
        self._project = project

    @classmethod
    def id_for(cls) -> str:
        """
        Get the job ID.
        """
        return "generate-entity-types-json"

    @override
    async def do(self, scheduler: Scheduler, /) -> None:
        await gather(
            *[
                scheduler.add(_GenerateEntityTypeJson(self._project, entity_type))
                for entity_type in await self._project.plugins.plugins(EntityDefinition)
            ]
        )


@final
class _GenerateEntityTypeJson(Job):
    def __init__(self, project: Project, entity_type: EntityDefinition):
        super().__init__(self.id_for(entity_type))
        self._project = project
        self._entity_type = entity_type

    @classmethod
    def id_for(cls, entity_type: EntityDefinition) -> str:
        return f"generate-entity-type-json:{entity_type.id}"

    @override
    async def do(self, scheduler: Scheduler, /) -> None:

        url_generator = await self._project.url_generator
        entity_type_path = self._project.www_directory / self._entity_type.id
        data: PortableMapping = {
            "$schema": await ProjectSchema.def_url(
                self._project,
                f"{kebab_case_to_lower_camel_case(self._entity_type.id)}EntityCollectionResponse",
            ),
            "collection": [],
        }
        for entity in self._project.ancestry[self._entity_type.cls]:
            cast("MutableSequence[str]", data["collection"]).append(
                url_generator.generate(
                    entity,
                    media_type=JSON,
                    absolute=True,
                )
            )
        rendered_json = dumps(data)
        async with create_json_resource(entity_type_path) as f:
            await f.write(rendered_json)


@final
class GenerateEntityTypesHtml(Job):
    """
    Generate HTML pages for entity types.
    """

    def __init__(self, *, per_page: int = 50, project: Project):
        super().__init__(self.id_for(), priority=True)
        self._project = project
        self._per_page = per_page

    @classmethod
    def id_for(cls) -> str:
        """
        Get the job ID.
        """
        return "generate-entity-types-html"

    @override
    async def do(self, scheduler: Scheduler, /) -> None:

        await gather(
            *[
                scheduler.add(
                    _GenerateEntityTypeHtml(
                        self._project,
                        entity_type,
                        locale,
                        page,
                        self._per_page,
                        page_count,
                    )
                )
                for entity_type in await self._project.plugins.plugins(EntityDefinition)
                if entity_type.public_facing
                and (
                    entity_type.id in self._project.configuration.entity_types
                    and self._project.configuration.entity_types[
                        entity_type.id
                    ].generate_html_list
                )
                and (
                    page_count := ceil(
                        len(self._project.ancestry[entity_type]) / self._per_page
                    )
                    # Always show at least the first page, even if there are no entities.
                    or 1
                )
                for page in range(page_count)
                for locale in self._project.configuration.locales.keys()  # noqa: SIM118
            ]
        )


@final
class _GenerateEntityTypeHtml(Job):
    def __init__(
        self,
        project: Project,
        entity_type: EntityDefinition,
        locale: Locale,
        page: int,
        per_page: int,
        page_count: int,
    ):
        super().__init__(self.id_for(entity_type, locale, page))
        self._project = project
        self._entity_type = entity_type
        self._locale = locale
        self._page = page
        self._per_page = per_page
        self._page_count = page_count

    @classmethod
    def id_for(cls, entity_type: EntityDefinition, locale: Locale, page: int) -> str:
        return f"generate-entity-type-html:{entity_type.id}:{locale}:{page}"

    @override
    async def do(self, scheduler: Scheduler, /) -> None:
        context = scheduler.context

        localizers = await self._project.localizers
        jinja = await self._project.jinja
        template = jinja.select_template(
            [
                f"entity/page-list--{self._entity_type.id}.html.j2",
                "entity/page-list.html.j2",
            ]
        )
        rendered_html = await template.render_async(
            document=await self._project.new_document(
                self._entity_type,
                self._entity_type,
                context=context,
                localizer=localizers.get(self._locale),
            ),
            page=self._page,
            per_page=self._per_page,
            page_count=self._page_count,
            page_entities=list(
                filter(is_public, self._project.ancestry[self._entity_type.cls])
            )[
                self._per_page * self._page : self._per_page * self._page
                + self._per_page
            ],
        )
        page_path = (
            self._project.localize_www_directory(self._locale) / self._entity_type.id
        )
        if self._page > 0:
            page_path /= f"page-{self._page + 1}"
        async with create_html_resource(page_path) as f:
            await f.write(rendered_html)


@final
class GenerateEntitiesJson(Job):
    """
    Generate JSON resources for entities.
    """

    def __init__(self, *, project: Project):
        super().__init__(self.id_for(), priority=True)
        self._project = project

    @classmethod
    def id_for(cls) -> str:
        """
        Get the job ID.
        """
        return "generate-entities-json"

    @override
    async def do(self, scheduler: Scheduler, /) -> None:

        await gather(
            *[
                scheduler.add(
                    _GenerateEntityJson(self._project, entity_type, entity.id)
                )
                for entity_type in await self._project.plugins.plugins(EntityDefinition)
                for entity in self._project.ancestry[entity_type.cls]
                if persistent_id(entity)
            ]
        )


@final
class _GenerateEntityJson(Job):
    def __init__(self, project: Project, entity_type: EntityDefinition, entity_id: str):
        super().__init__(self.id_for(entity_type, entity_id))
        self._project = project
        self._entity_type = entity_type
        self._entity_id = entity_id

    @classmethod
    def id_for(cls, entity_type: EntityDefinition, entity_id: str) -> str:
        return f"generate-entity-json:{entity_type.id}:{entity_id}"

    @override
    async def do(self, scheduler: Scheduler, /) -> None:

        entity = self._project.ancestry[self._entity_type.cls][self._entity_id]
        entity_path = (
            self._project.www_directory / self._entity_type.id / entity.public_id
        )
        rendered_json = dumps(await entity.dump_linked_data(self._project))
        async with create_json_resource(entity_path) as f:
            await f.write(rendered_json)


@final
class GenerateEntitiesHtml(Job):
    """
    Generate HTML pages for entities.
    """

    def __init__(self, *, project: Project):
        super().__init__(self.id_for(), priority=True)
        self._project = project

    @classmethod
    def id_for(cls) -> str:
        """
        Get the job ID.
        """
        return "generate-entities-html"

    @override
    async def do(self, scheduler: Scheduler, /) -> None:

        await gather(
            *[
                scheduler.add(
                    _GenerateEntityHtml(self._project, entity_type, entity.id, locale)
                )
                for entity_type in await self._project.plugins.plugins(EntityDefinition)
                if entity_type.public_facing
                for entity in self._project.ancestry[entity_type.cls]
                if persistent_id(entity) and is_public(entity)
                for locale in self._project.configuration.locales.keys()  # noqa: SIM118
            ]
        )


@final
class _GenerateEntityHtml(Job):
    def __init__(
        self,
        project: Project,
        entity_type: EntityDefinition,
        entity_id: str,
        locale: Locale,
    ):
        super().__init__(self.id_for(entity_type, entity_id, locale))
        self._project = project
        self._entity_type = entity_type
        self._entity_id = entity_id
        self._locale = locale

    @classmethod
    def id_for(
        cls, entity_type: EntityDefinition, entity_id: str, locale: Locale
    ) -> str:
        return f"generate-entity-html:{entity_type.id}:{entity_id}:{locale}"

    @override
    async def do(self, scheduler: Scheduler, /) -> None:
        context = scheduler.context

        localizers = await self._project.localizers
        jinja = await self._project.jinja
        entity = self._project.ancestry[self._entity_type.cls][self._entity_id]
        entity_path = (
            self._project.localize_www_directory(self._locale)
            / self._entity_type.id
            / entity.public_id
        )
        rendered_html = await jinja.select_template(
            [
                f"entity/page--{self._entity_type.id}.html.j2",
                "entity/page.html.j2",
            ]
        ).render_async(
            document=await self._project.new_document(
                entity,
                entity,
                context=context,
                localizer=localizers.get(self._locale),
            )
        )
        async with create_html_resource(entity_path) as f:
            await f.write(rendered_html)



"""
Sequence data types.
"""

from __future__ import annotations

from collections.abc import Callable, MutableSequence
from typing import TYPE_CHECKING, Any

from betty.data.aggregate.collection import CollectionDefinition
from betty.data.indicator.selector import Index
from betty.portable import CallbackPorter, PortableData

if TYPE_CHECKING:
    from ty_extensions import Intersection

    from betty.data import Data, DataDefinition
    from betty.locale.localizable import ResolvableLocalizable


class SequenceDefinition[MutableSequenceT: MutableSequence[Any]](
    CollectionDefinition[MutableSequenceT, Index]
):
    """
    A sequence data definition.
    """

    def __init__[ValueT](
        self,
        /,
        cls: type[Intersection[MutableSequenceT, MutableSequence[ValueT]]]
        | None = None,
        *,
        value: DataDefinition[ValueT] | type[Intersection[ValueT, Data]],
        label: ResolvableLocalizable,
        description: ResolvableLocalizable | None = None,
        factory: Callable[[], MutableSequenceT] | None = None,
    ):
        super().__init__(
            cls=cls,
            item=value,
            label=label,
            description=description,
            porter=CallbackPorter(self._load, self._dump),
        )
        self._factory = factory

    def _load(self, portable: PortableData, /) -> MutableSequenceT:
        from betty.assertion import assert_sequence

        loaded = self.new()
        loaded.extend(assert_sequence(self._item.porter.load)(portable))
        return loaded

    def _dump(self, data: MutableSequenceT) -> PortableData:
        return [self._item.porter.dump(item) for item in data]



"""
Plain localizables.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, final, override

from betty.locale import HasLocale, HasLocaleStr, ResolvableLocale, resolve_locale
from betty.locale.localizable import Localizable

if TYPE_CHECKING:
    from babel import Locale
    from ty_extensions import Intersection

    from betty.locale.localize import Localizer


@final
class Plain(Localizable):
    """
    Turns a plain string into a :py:class:`betty.locale.localizable.Localizable` without any actual translations.
    """

    def __init__(self, text: str, locale: ResolvableLocale | None = None, /):
        from betty.assertion import assert_str

        assert_str(minimum_length=1)(text)
        self._text = text
        self._locale = None if locale is None else resolve_locale(locale)

    @property
    def text(self) -> str:
        """
        The plain text.
        """
        return self._text

    @property
    def locale(self) -> Locale | None:
        """
        The locale the text is in.
        """
        return self._locale

    @override
    def localize(self, localizer: Localizer, /) -> Intersection[HasLocale, str]:
        return HasLocaleStr(self._text, locale=self._locale)



"""
Populate ancestries with information from Wikipedia and Wikimedia.
"""

from __future__ import annotations

from asyncio import gather
from collections import defaultdict
from typing import TYPE_CHECKING

from betty.ancestry.file import File
from betty.ancestry.file_reference import FileReference
from betty.ancestry.has_file_references import HasFileReferences
from betty.ancestry.has_links import HasLinks
from betty.ancestry.link import Link
from betty.ancestry.place import Place
from betty.concurrent import AsynchronizedLock, Lock
from betty.functools import map_suppress
from betty.locale import negotiate_locale, resolve_locale
from betty.locale.error import LocaleError
from betty.locale.localizable.gettext import _
from betty.locale.localizable.static import StaticTranslations
from betty.media_type.media_types import HTML
from betty.typing import threadsafe
from betty.wiki import NotAPageError, parse_page_link
from betty.wiki.client import ClientError

if TYPE_CHECKING:
    from collections.abc import Mapping, MutableMapping, Sequence

    from babel import Locale
    from ty_extensions import Intersection

    from betty.ancestry import Ancestry
    from betty.copyright_notice import CopyrightNotice
    from betty.locale.localize import LocalizerRepository
    from betty.model import Entity
    from betty.user import User
    from betty.wiki.client import Client, Image


@threadsafe
class Populator:
    """
    Populate an ancestry with information from Wikipedia and Wikimedia.
    """

    def __init__(
        self,
        ancestry: Ancestry,
        locales: Sequence[Locale],
        localizers: LocalizerRepository,
        client: Client,
        copyright_notice: CopyrightNotice,
        *,
        user: User,
    ):
        self._ancestry = ancestry
        self._locales = locales
        self._localizers = localizers
        self._client = client
        self._image_files: MutableMapping[Image, File] = {}
        self._image_files_locks: Mapping[Image, Lock] = defaultdict(
            AsynchronizedLock.new_threadsafe
        )
        self._copyright_notice = copyright_notice
        self._user = user

    async def populate(self, entity: Entity) -> None:
        """
        Populate an entity.
        """
        populations = []
        if isinstance(entity, HasFileReferences) and isinstance(entity, HasLinks):
            populations.append(self._populate_has_file_references_and_links(entity))
        if isinstance(entity, Place):
            populations.append(self._populate_place(entity))
        if isinstance(entity, Link):
            populations.append(self._populate_link(entity))
        await gather(*populations)

    async def _populate_link(self, link: Link) -> None:
        try:
            page_language, page_name = parse_page_link(
                link, [self._localizers.get(locale) for locale in self._locales]
            )
        except NotAPageError:
            return
        if link.media_type is None:
            link.media_type = HTML
        if link.relationship is None:
            link.relationship = "external"
        if link.description is None:
            link.description = _("Read more on Wikipedia.")
        try:
            page_translations = dict(
                await self._client.get_translations(page_language, page_name)
            )
        except ClientError as error:
            await self._user.message_warning(error)
            return
        if page_translations:
            # For convenience, we add the original page language and name to the available translations.
            page_translations[page_language] = page_name

            # Most Wikipedia languages are based on ISO 639-1 and ISO 639-3
            # (https://en.wikipedia.org/wiki/List_of_Wikipedias). However, some languages such as "simple" are not.
            translation_page_locales = list(
                map_suppress(resolve_locale, LocaleError, page_translations)
            )

            locales_to_page_languages = {}
            for locale in self._locales:
                negotiated_page_language = negotiate_locale(
                    locale, translation_page_locales
                )
                locales_to_page_languages[locale] = (
                    page_language
                    if negotiated_page_language is None
                    else str(negotiated_page_language)
                )

            link.url = StaticTranslations(
                {
                    locale: f"https://{page_language}.wikipedia.org/wiki/{page_translations[page_language]}"
                    for locale, page_language in locales_to_page_languages.items()
                }
            )
            if not link.has_label:
                link.label = StaticTranslations(
                    {
                        locale: await self._fetch_link_label_from_page(
                            page_language, page_translations[page_language]
                        )
                        or page_name
                        for locale, page_language in locales_to_page_languages.items()
                    }
                )

    async def _fetch_link_label_from_page(
        self, page_language: str, page_name: str
    ) -> str | None:
        try:
            summary = await self._client.get_summary(page_language, page_name)
        except ClientError as error:
            await self._user.message_warning(error)
            return None
        return summary.title

    async def _populate_place(self, place: Place) -> None:
        await self._populate_place_coordinates(place)

    async def _populate_place_coordinates(self, place: Place) -> None:
        await gather(
            *(
                self._populate_place_coordinates_link(place, link)
                for link in place.links
            )
        )

    async def _populate_place_coordinates_link(self, place: Place, link: Link) -> None:
        try:
            page_language, page_name = parse_page_link(
                link, [self._localizers.get(locale) for locale in self._locales]
            )
        except NotAPageError:
            return
        else:
            try:
                coordinates = await self._client.get_place_coordinates(
                    page_language, page_name
                )
            except ClientError as error:
                await self._user.message_warning(error)
                return
            if coordinates:
                place.coordinates = coordinates

    async def _populate_has_file_references_and_links(
        self, has_file_references: Intersection[HasFileReferences, HasLinks]
    ) -> None:
        await gather(
            *(
                self._populate_has_file_references_and_links_link(
                    has_file_references, link
                )
                for link in has_file_references.links
            )
        )

    async def _populate_has_file_references_and_links_link(
        self, has_file_references: HasFileReferences, link: Link
    ) -> None:
        try:
            page_language, page_name = parse_page_link(
                link, [self._localizers.get(locale) for locale in self._locales]
            )
        except NotAPageError:
            return
        else:
            try:
                image = await self._client.get_image(page_language, page_name)
            except ClientError as error:
                await self._user.message_warning(error)
                return
            if not image:
                return
            await self._image_file_reference(has_file_references, image)

    async def _image_file_reference(
        self, has_file_references: HasFileReferences, image: Image
    ) -> FileReference:
        async with self._image_files_locks[image]:
            try:
                file = self._image_files[image]
            except KeyError:
                file = File(
                    id=f"wikipedia-{image.title}",
                    name=image.name,
                    path=image.path,
                    media_type=image.media_type,
                    links=[
                        Link(
                            StaticTranslations(
                                {
                                    locale: f"{image.wikimedia_commons_url}?uselang={locale}"
                                    for locale in self._locales
                                }
                            ),
                            label=_("Description, licensing, and image history"),
                            description=_(
                                "Find out more about this image on Wikimedia Commons."
                            ),
                            media_type=HTML,
                        )
                    ],
                    copyright_notice=self._copyright_notice,
                )
                self._image_files[image] = file
                self._ancestry.add(file)
            file_reference = FileReference(has_file_references, file)
            self._ancestry.add(file_reference)
            return file_reference



"""
Provide JSON utilities.
"""

from __future__ import annotations

import enum
from typing import TYPE_CHECKING, Any, cast, final, override

from jsonschema.validators import Draft202012Validator
from referencing import Registry, Resource

from betty.classtools import Singleton
from betty.locale.localize import DEFAULT_LOCALIZER, resolve_localized
from betty.portable import PortableData, PortableMapping

if TYPE_CHECKING:
    from betty.locale.localizable import ResolvableLocalizable


class Schema:
    """
    A JSON Schema.

    All schemas using this class **MUST** follow JSON Schema Draft 2020-12.

    To test your own subclasses, use :py:class:`betty.test_utils.json.schema.SchemaTestBase`.
    """

    def __init__(
        self,
        *,
        def_name: str | None = None,
        title: ResolvableLocalizable | None = None,
        description: ResolvableLocalizable | None = None,
    ):
        self._def_name = def_name
        self._schema: PortableMapping = {
            # The entire API assumes this dialect, so enforce it.
            "$schema": "https://json-schema.org/draft/2020-12/schema",
        }
        if title is not None:
            self.title = title
        if description is not None:
            self.description = description

    @property
    def def_name(self) -> str | None:
        """
        The schema machine name when embedded into another schema's ``$defs``.
        """
        return self._def_name

    @property
    def schema(self) -> PortableMapping:
        """
        The raw JSON Schema.
        """
        return self._schema

    @property
    def title(self) -> str | None:
        """
        The schema's human-readable US English (short) title.
        """
        try:
            return cast(str, self._schema["title"])
        except KeyError:
            return None

    @title.setter
    def title(self, title: ResolvableLocalizable) -> None:
        self._schema["title"] = resolve_localized(title, localizer=DEFAULT_LOCALIZER)

    @property
    def description(self) -> str | None:
        """
        The schema's human-readable US English (long) description.
        """
        try:
            return cast(str, self._schema["description"])
        except KeyError:
            return None

    @description.setter
    def description(self, description: ResolvableLocalizable) -> None:
        self._schema["description"] = resolve_localized(
            description, localizer=DEFAULT_LOCALIZER
        )

    @property
    def defs(self) -> PortableMapping:
        """
        The JSON Schema's ``$defs`` definitions, kept separately, so they can be merged when this schema is embedded.

        Only top-level definitions are supported. You **MUST NOT** nest definitions. Instead, prefix or suffix
        their names.
        """
        return cast(PortableMapping, self._schema.setdefault("$defs", {}))

    def embed(self, into: Schema, /) -> PortableMapping:
        """
        Embed this schema.

        This is where the raw schema may be enhanced before being returned.
        """
        for name, schema in self.defs.items():
            into.defs[name] = schema
        schema = {
            child_name: child_schema
            for child_name, child_schema in self.schema.items()
            if child_name not in ("$defs", "$schema")
        }
        if self._def_name is None:
            return schema
        into.defs[self._def_name] = schema
        return Ref(self._def_name).embed(into)

    def validate(self, data: Any, /) -> None:
        """
        Validate data against this schema.
        """
        schema = self.schema
        if "$id" not in schema:
            schema["$id"] = "https://betty.example.com"
        schema_registry = Resource.from_contents(schema) @ Registry()
        validator = Draft202012Validator(
            schema,
            registry=schema_registry,
        )
        validator.validate(data)


class _Type(Schema):
    _type: str

    def __init__(
        self,
        *,
        def_name: str | None = None,
        title: ResolvableLocalizable | None = None,
        description: ResolvableLocalizable | None = None,
    ):
        super().__init__(def_name=def_name, title=title, description=description)
        self._schema["type"] = self._type


class String(_Type):
    """
    A JSON Schema ``string`` type.
    """

    _type = "string"

    class Format(enum.Enum):
        """
        A JSON Schema ``string`` type's ``format``.
        """

        DATE_TIME = "date-time"
        TIME = "time"
        DATE = "date"
        DURATION = "duration"
        EMAIL = "email"
        IDN_EMAIL = "idn-email"
        HOSTNAME = "hostname"
        IDN_HOSTNAME = "idn-hostname"
        IPV4 = "ipv4"
        IPV6 = "ipv6"
        UUID = "uuid"
        URI = "uri"
        URI_REFERENCE = "uri-reference"
        IRI = "iri"
        IRI_REFERENCE = "iri-reference"
        URI_TEMPLATE = "uri-template"
        JSON_POINTER = "json-pointer"
        RELATIVE_JSON_POINTER = "relative-json-pointer"
        REGEX = "regex"

    def __init__(
        self,
        *,
        def_name: str | None = None,
        title: ResolvableLocalizable | None = None,
        description: ResolvableLocalizable | None = None,
        min_length: int | None = None,
        max_length: int | None = None,
        pattern: str | None = None,
        format: Format | None = None,  # noqa: A002
    ):
        super().__init__(
            def_name=def_name,
            title=title,
            description=description,
        )
        if min_length is not None:
            self._schema["minLength"] = min_length
        if max_length is not None:
            self._schema["maxLength"] = max_length
        if pattern is not None:
            self._schema["pattern"] = pattern
        if format is not None:
            self._schema["format"] = format.value


class Boolean(_Type):
    """
    A JSON Schema ``boolean`` type.
    """

    _type = "boolean"


class Number(_Type):
    """
    A JSON Schema ``number`` type.
    """

    _type = "number"


class Integer(_Type):
    """
    A JSON Schema ``integer`` type.
    """

    _type = "integer"


class Null(_Type):
    """
    A JSON Schema ``null`` type.
    """

    _type = "null"


class Object(_Type):
    """
    A JSON Schema ``object`` type.
    """

    _type = "object"

    def __init__(
        self,
        *,
        def_name: str | None = None,
        title: ResolvableLocalizable | None = None,
        description: ResolvableLocalizable | None = None,
    ):
        super().__init__(
            def_name=def_name,
            title=title,
            description=description,
        )
        self._properties = self._schema["properties"] = {}
        self._required = self._schema["required"] = []

    def add_property(
        self,
        property_name: str,
        property_schema: Schema,
        property_required: bool = True,
        /,
    ) -> None:
        """
        Add a property to the object schema.
        """
        self._properties[property_name] = property_schema.embed(self)
        if property_required:
            self._required.append(property_name)


class Array(_Type):
    """
    A JSON Schema ``array`` type.
    """

    _type = "array"

    def __init__(
        self,
        items: Schema,
        *,
        def_name: str | None = None,
        title: ResolvableLocalizable | None = None,
        description: ResolvableLocalizable | None = None,
    ):
        super().__init__(
            def_name=def_name,
            title=title,
            description=description,
        )
        self._schema["items"] = items.embed(self)


class _Container(Schema):
    _type: str

    def __init__(
        self,
        *items: Schema,
        def_name: str | None = None,
        title: ResolvableLocalizable | None = None,
        description: ResolvableLocalizable | None = None,
    ):
        super().__init__(def_name=def_name, title=title, description=description)
        self._schema[self._type] = [item.embed(self) for item in items]


class AllOf(_Container):
    """
    A JSON Schema ``allOf``.
    """

    _type = "allOf"


class AnyOf(_Container):
    """
    A JSON Schema ``anyOf``.
    """

    _type = "anyOf"


class OneOf(_Container):
    """
    A JSON Schema ``oneOf``.
    """

    _type = "oneOf"


class Const(Schema):
    """
    A JSON Schema ``const``.
    """

    def __init__(
        self,
        const: PortableData,
        *,
        def_name: str | None = None,
        title: ResolvableLocalizable | None = None,
        description: ResolvableLocalizable | None = None,
    ):
        super().__init__(def_name=def_name, title=title, description=description)
        self._schema["const"] = const


class Enum(Schema):
    """
    A JSON Schema ``enum``.
    """

    def __init__(
        self,
        *values: PortableData,
        def_name: str | None = None,
        title: ResolvableLocalizable | None = None,
        description: ResolvableLocalizable | None = None,
    ):
        super().__init__(def_name=def_name, title=title, description=description)
        self._schema["enum"] = list(values)


class Def(str):
    """
    The name of a named Betty schema.

    Using this instead of :py:class:`str` directly allows Betty to
    bundle schemas together under a project namespace.

    See :py:attr:`betty.json.schema.Schema.def_name`.
    """

    __slots__ = ()

    @override
    def __new__(cls, def_name: str, /):
        return super().__new__(cls, f"#/$defs/{def_name}")


class Ref(Schema):
    """
    A JSON Schema that references a named Betty schema.
    """

    def __init__(self, def_name: str, /):
        super().__init__()
        self._schema["$ref"] = Def(def_name)


class JsonSchemaReference(String):
    """
    The JSON Schema schema.
    """

    def __init__(self):
        super().__init__(
            def_name="jsonSchemaReference",
            title="JSON Schema reference",
            format=String.Format.URI,
            description="A JSON Schema URI.",
        )


@final
class JsonSchemaSchema(Singleton, Schema):
    """
    The JSON Schema Draft 2020-12 schema.
    """

    _SCHEMA: PortableMapping = {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "$id": "https://json-schema.org/draft/2020-12/schema",
        "$vocabulary": {
            "https://json-schema.org/draft/2020-12/vocab/core": True,
            "https://json-schema.org/draft/2020-12/vocab/applicator": True,
            "https://json-schema.org/draft/2020-12/vocab/unevaluated": True,
            "https://json-schema.org/draft/2020-12/vocab/validation": True,
            "https://json-schema.org/draft/2020-12/vocab/meta-data": True,
            "https://json-schema.org/draft/2020-12/vocab/format-annotation": True,
            "https://json-schema.org/draft/2020-12/vocab/content": True,
        },
        "$dynamicAnchor": "meta",
        "title": "Core and Validation specifications meta-schema",
        "allOf": [
            {"$ref": "meta/core"},
            {"$ref": "meta/applicator"},
            {"$ref": "meta/unevaluated"},
            {"$ref": "meta/validation"},
            {"$ref": "meta/meta-data"},
            {"$ref": "meta/format-annotation"},
            {"$ref": "meta/content"},
        ],
        "type": ["object", "boolean"],
        "$comment": "This meta-schema also defines keywords that have appeared in previous drafts in order to prevent incompatible extensions as they remain in common use.",
        "properties": {
            "definitions": {
                "$comment": '"definitions" has been replaced by "$defs".',
                "type": "object",
                "additionalProperties": {"$dynamicRef": "#meta"},
                "deprecated": True,
                "default": {},
            },
            "dependencies": {
                "$comment": '"dependencies" has been split and replaced by "dependentSchemas" and "dependentRequired" in order to serve their differing semantics.',
                "type": "object",
                "additionalProperties": {
                    "anyOf": [
                        {"$dynamicRef": "#meta"},
                        {"$ref": "meta/validation#/$defs/stringArray"},
                    ]
                },
                "deprecated": True,
                "default": {},
            },
            "$recursiveAnchor": {
                "$comment": '"$recursiveAnchor" has been replaced by "$dynamicAnchor".',
                "$ref": "meta/core#/$defs/anchorString",
                "deprecated": True,
            },
            "$recursiveRef": {
                "$comment": '"$recursiveRef" has been replaced by "$dynamicRef".',
                "$ref": "meta/core#/$defs/uriReferenceString",
                "deprecated": True,
            },
        },
    }

    def __init__(self):
        super().__init__(def_name="jsonSchema", title="JSON Schema")
        self._schema = self._SCHEMA



from __future__ import annotations  # noqa: D100

from typing import TYPE_CHECKING, Self, final, override

from betty.argparse import assertion_to_argument_type
from betty.assertion import assert_locale
from betty.console.command import Command, CommandDefinition, CommandFunction
from betty.extension import ExtensionDefinition
from betty.locale.localizable.gettext import _
from betty.locale.translation.project import extension as extension_translation
from betty.locale.translation.project import extension as translation_project_extension
from betty.service.factory import Manufacturable
from betty.service.requirement.app import require_app

if TYPE_CHECKING:
    import argparse

    from babel import Locale

    from betty.app import App


@final
@CommandDefinition(
    "extension-new-translation", label=_("Create a new translation for an extension")
)
class ExtensionNewTranslation(Manufacturable, Command):
    """
    .. plugin:: command:extension-new-translation.
    """

    def __init__(self, app: App):
        self._app = app

    @override
    @classmethod
    @require_app
    async def new(cls, app: App, /) -> Self:
        return cls(app)

    @override
    async def configure(self, parser: argparse.ArgumentParser) -> CommandFunction:
        extensions = await self._app.plugins.plugins(ExtensionDefinition)
        localizer = await self._app.localizer
        parser.add_argument(
            "extension",
            type=assertion_to_argument_type(
                lambda extension_id: (
                    translation_project_extension.assert_extension_has_assets_directory_path(
                        extensions[extension_id]
                    )
                ),
                localizer=localizer,
            ),
        )
        parser.add_argument(
            "locale",
            type=assertion_to_argument_type(assert_locale(), localizer=localizer),
        )
        return self._command_function

    async def _command_function(
        self, extension: ExtensionDefinition, locale: Locale
    ) -> None:
        await extension_translation.new_extension_translation(
            locale, extension, user=self._app.user
        )



"""
Provide plugin assertions.
"""

from typing import Any

from betty.assertion import AssertionChain, assert_str
from betty.exception import HumanFacingException
from betty.locale.localizable.gettext import _
from betty.locale.localizable.markup import Paragraph, do_you_mean
from betty.plugin import PluginDefinition
from betty.plugin.error import PluginNotFound
from betty.plugin.repository import PluginRepository


def assert_plugin[PluginDefinitionT: PluginDefinition](
    plugins: PluginRepository[PluginDefinitionT],
) -> AssertionChain[Any, PluginDefinitionT]:
    """
    Assert that a value is a plugin ID.
    """

    def _assert(
        value: Any,
    ) -> PluginDefinitionT:
        plugin_id = assert_str()(value)
        try:
            return plugins[plugin_id]
        except PluginNotFound:
            raise HumanFacingException(
                Paragraph(
                    _('Unknown plugin "{plugin_id}".').format(plugin_id=plugin_id),
                    do_you_mean(*(f'"{plugin.id}"' for plugin in plugins)),
                )
            ) from None

    return AssertionChain(_assert)



"""
Provide Betty's ancestry event types.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, final, override

from betty.definition.human_facing import CountableHumanFacingDefinition
from betty.locale.localizable.gettext import _, ngettext
from betty.plugin import Plugin, PluginTypeDefinition, ResolvableId, resolve_id
from betty.plugin.discovery.entry_point import EntryPointDiscovery
from betty.plugin.factory import PluginManufacturer
from betty.plugin.ordered import OrderedPluginDefinition
from betty.service.requirement.project import require_project

if TYPE_CHECKING:
    import builtins
    from collections.abc import Set

    from betty.ancestry.person import Person
    from betty.locale.localizable import CountableLocalizable, ResolvableLocalizable
    from betty.machine_name import MachineName, ResolvableMachineName
    from betty.project import Project


class EventType(Plugin["EventTypeDefinition"]):
    """
    Define an :py:class:`betty.ancestry.event.Event` type.
    """


class ShouldExistEventType(EventType, ABC):
    """
    An event type that controls whether at least one event of this type should exist for a person.
    """

    @classmethod
    @abstractmethod
    async def should_exist(cls, project: Project, person: Person) -> bool:
        """
        Whether at least one event of this type should exist for the given person.
        """


@final
@PluginTypeDefinition(
    "event-type",
    label=_("Event type"),
    label_plural=_("Event types"),
    label_countable=ngettext("{count} event type", "{count} event types"),
    discovery=[
        EntryPointDiscovery("betty.event_type"),
        require_project(
            lambda project: (
                configuration.new_plugin()
                for configuration in project.configuration.event_types
            )
        ),
    ],
)
class EventTypeDefinition(
    CountableHumanFacingDefinition, OrderedPluginDefinition[EventType]
):
    """
    .. plugin_type:: event-type.
    """

    def __init__(
        self,
        plugin_id: ResolvableMachineName,
        *,
        label: ResolvableLocalizable,
        label_plural: ResolvableLocalizable,
        label_countable: CountableLocalizable,
        description: ResolvableLocalizable | None = None,
        comes_before: Set[ResolvableId] | None = None,
        comes_after: Set[ResolvableId] | None = None,
        indicates: ResolvableId[EventTypeDefinition] | None = None,
    ):
        super().__init__(
            plugin_id,
            label=label,
            label_plural=label_plural,
            label_countable=label_countable,
            description=description,
            comes_before=comes_before,
            comes_after=comes_after,
        )
        self._indicates = None if indicates is None else resolve_id(indicates)

    @property
    def indicates(self) -> MachineName | None:
        """
        Return whether events of this type (approximately) indicate that an event of the retuned type has happened.
        """
        return self._indicates


@final
class EventTypeManufacturer(PluginManufacturer[EventTypeDefinition, EventType]):
    """
    The event type manufacturer.
    """

    @override
    @classmethod
    def type(cls) -> builtins.type[EventTypeDefinition]:
        return EventTypeDefinition



"""
Mapping types and implementations.
"""

from abc import abstractmethod
from collections.abc import Iterable, Mapping, MutableMapping
from typing import overload, override

from betty.collection import MutableCollection


class ResolvedMapping[KeyT, ResolvableKeyT, ValueT](Mapping[KeyT, ValueT]):
    """
    A mutable mapping of resolvable keys.
    """

    @override
    @abstractmethod
    def __getitem__(self, key: KeyT | ResolvableKeyT) -> ValueT:
        pass

    @overload
    def get[T](self, key: KeyT | ResolvableKeyT, default: T, /) -> ValueT | T:
        pass

    @overload
    def get(self, key: KeyT | ResolvableKeyT, default: None = None, /) -> ValueT | None:
        pass

    @override
    @abstractmethod
    def get[T](
        self, key: KeyT | ResolvableKeyT, default: T | None = None, /
    ) -> ValueT | None:
        pass


class MutableResolvedMapping[KeyT, ResolvableKeyT, ValueT, ResolvableValueT](
    MutableMapping[KeyT, ValueT],
    MutableCollection[KeyT],
    ResolvedMapping[KeyT, ResolvableKeyT, ValueT],
):
    """
    A mutable mapping of resolvable keys and values.
    """

    @abstractmethod
    def __setitem__(
        self, key: KeyT | ResolvableKeyT, value: ValueT | ResolvableValueT
    ) -> None:
        pass

    @abstractmethod
    def __delitem__(self, key: KeyT | ResolvableKeyT) -> None:
        pass

    @overload
    def update(
        self,
        other: Mapping[KeyT | ResolvableKeyT, ValueT | ResolvableValueT]
        | Iterable[tuple[KeyT | ResolvableKeyT, ValueT | ResolvableValueT]],
        /,
        **kwargs: ValueT | ResolvableValueT,
    ) -> None:
        pass

    @overload
    def update(self, **kwargs: ValueT | ResolvableValueT) -> None:
        pass

    @override
    @abstractmethod
    def update(self, other=None, **kwargs) -> None:  # ty:ignore[invalid-method-override]
        pass

    @overload
    def setdefault[T](
        self: MutableMapping[KeyT, T | None],
        key: KeyT | ResolvableKeyT,
        default: None = None,
        /,
    ) -> T | None:
        pass

    @overload
    def setdefault[T](
        self, key: KeyT | ResolvableKeyT, default: ValueT | ResolvableValueT, /
    ) -> ValueT:
        pass

    @override
    @abstractmethod
    def setdefault(self, key, default):
        pass

    @overload
    def pop(self, key: KeyT | ResolvableKeyT, /) -> ValueT:
        pass

    @overload
    def pop(
        self, key: KeyT | ResolvableKeyT, /, default: ValueT | ResolvableValueT
    ) -> ValueT:
        pass

    @overload
    def pop[T](self, key: KeyT | ResolvableKeyT, /, default: T) -> ValueT | T:
        pass

    @override
    @abstractmethod
    def pop(self, key, default):
        pass

    @override
    @abstractmethod
    def popitem(self) -> tuple[KeyT, ValueT]:
        pass
